# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import random

from copy import deepcopy
from typing import Dict

from rest_framework.decorators import api_view
from rest_framework.response import Response

from swh.indexer.fossology_license import FossologyLicenseIndexer
from swh.indexer.mimetype import MimetypeIndexer
from swh.indexer.ctags import CtagsIndexer
from swh.indexer.storage import get_indexer_storage
from swh.model.from_disk import Directory
from swh.model.hashutil import hash_to_hex, hash_to_bytes, DEFAULT_ALGORITHMS
from swh.model.identifiers import directory_identifier
from swh.loader.git.from_disk import GitLoaderFromArchive
from swh.search import get_search
from swh.storage.algos.dir_iterators import dir_iterator
from swh.web import config
from swh.web.browse.utils import (
    get_mimetype_and_encoding_for_content, prepare_content_for_display,
    _re_encode_content
)
from swh.web.common import service
from swh.web.common.highlightjs import get_hljs_language_from_filename

# Module used to initialize data that will be provided as tests input

# Configuration for git loader
_TEST_LOADER_CONFIG = {
    'storage': {
        'cls': 'memory'
    },
    'save_data': False,
    'max_content_size': 100 * 1024 * 1024,
}

# Base content indexer configuration
_TEST_INDEXER_BASE_CONFIG = {
    'storage': {
        'cls': 'memory'
    },
    'objstorage': {
        'cls': 'memory',
        'args': {},
    },
    'indexer_storage': {
        'cls': 'memory',
        'args': {},
    }
}


def random_sha1():
    return hash_to_hex(bytes(random.randint(0, 255) for _ in range(20)))


def random_sha256():
    return hash_to_hex(bytes(random.randint(0, 255) for _ in range(32)))


def random_blake2s256():
    return hash_to_hex(bytes(random.randint(0, 255) for _ in range(32)))


def random_content():
    return {
        'sha1': random_sha1(),
        'sha1_git': random_sha1(),
        'sha256': random_sha256(),
        'blake2s256': random_blake2s256(),
    }


# MimetypeIndexer with custom configuration for tests
class _MimetypeIndexer(MimetypeIndexer):
    def parse_config_file(self, *args, **kwargs):
        return {
            **_TEST_INDEXER_BASE_CONFIG,
            'tools': {
                'name': 'file',
                'version': '1:5.30-1+deb9u1',
                'configuration': {
                    "type": "library",
                    "debian-package": "python3-magic"
                }
            }
        }


# FossologyLicenseIndexer with custom configuration for tests
class _FossologyLicenseIndexer(FossologyLicenseIndexer):
    def parse_config_file(self, *args, **kwargs):
        return {
            **_TEST_INDEXER_BASE_CONFIG,
            'workdir': '/tmp/swh/indexer.fossology.license',
            'tools': {
                'name': 'nomos',
                'version': '3.1.0rc2-31-ga2cbb8c',
                'configuration': {
                    'command_line': 'nomossa <filepath>',
                },
            }
        }


# CtagsIndexer with custom configuration for tests
class _CtagsIndexer(CtagsIndexer):
    def parse_config_file(self, *args, **kwargs):
        return {
            **_TEST_INDEXER_BASE_CONFIG,
            'workdir': '/tmp/swh/indexer.ctags',
            'languages': {'c': 'c'},
            'tools': {
                'name': 'universal-ctags',
                'version': '~git7859817b',
                'configuration': {
                    'command_line': '''ctags --fields=+lnz --sort=no --links=no ''' # noqa
                                    '''--output-format=json <filepath>'''
                },
            }
        }


# Lightweight git repositories that will be loaded to generate
# input data for tests
_TEST_ORIGINS = [
    {
        'type': 'git',
        'url': 'https://github.com/wcoder/highlightjs-line-numbers.js',
        'archives': ['highlightjs-line-numbers.js.zip',
                     'highlightjs-line-numbers.js_visit2.zip'],
        'visit_date': ['Dec 1 2018, 01:00 UTC',
                       'Jan 20 2019, 15:00 UTC']
    },
    {
        'type': 'git',
        'url': 'https://github.com/memononen/libtess2',
        'archives': ['libtess2.zip'],
        'visit_date': ['May 25 2018, 01:00 UTC']
    },
    {
        'type': 'git',
        'url': 'repo_with_submodules',
        'archives': ['repo_with_submodules.tgz'],
        'visit_date': ['Jan 1 2019, 01:00 UTC']
    }
]

_contents = {}


# Tests data initialization
def _init_tests_data():
    # To hold reference to the memory storage
    storage = None

    # Create search instance
    search = get_search('memory', {})
    search.initialize()
    search.origin_update({'url': origin['url']} for origin in _TEST_ORIGINS)

    # Load git repositories from archives
    for origin in _TEST_ORIGINS:
        for i, archive in enumerate(origin['archives']):
            origin_repo_archive = \
                os.path.join(os.path.dirname(__file__),
                             'resources/repos/%s' % archive)
            loader = GitLoaderFromArchive(origin['url'],
                                          archive_path=origin_repo_archive,
                                          config=_TEST_LOADER_CONFIG,
                                          visit_date=origin['visit_date'][i])
            if storage is None:
                storage = loader.storage
            else:
                loader.storage = storage
            loader.load()

        origin.update(storage.origin_get(origin))  # add an 'id' key if enabled
        search.origin_update([{'url': origin['url'], 'has_visits': True}])

    for i in range(250):
        url = 'https://many.origins/%d' % (i+1)
        storage.origin_add([{'url': url}])
        search.origin_update([{'url': url, 'has_visits': True}])
        visit = storage.origin_visit_add(url, '2019-12-03 13:55:05', 'tar')
        storage.origin_visit_update(
            url, visit['visit'],
            snapshot='1a8893e6a86f444e8be8e7bda6cb34fb1735a00e')

    contents = set()
    directories = set()
    revisions = set()
    releases = set()
    snapshots = set()

    content_path = {}

    # Get all objects loaded into the test archive
    for origin in _TEST_ORIGINS:
        snp = storage.snapshot_get_latest(origin['url'])
        snapshots.add(hash_to_hex(snp['id']))
        for branch_name, branch_data in snp['branches'].items():
            if branch_data['target_type'] == 'revision':
                revisions.add(branch_data['target'])
            elif branch_data['target_type'] == 'release':
                release = next(storage.release_get([branch_data['target']]))
                revisions.add(release['target'])
                releases.add(hash_to_hex(branch_data['target']))

        for rev_log in storage.revision_shortlog(set(revisions)):
            rev_id = rev_log[0]
            revisions.add(rev_id)

        for rev in storage.revision_get(revisions):
            dir_id = rev['directory']
            directories.add(hash_to_hex(dir_id))
            for entry in dir_iterator(storage, dir_id):
                content_path[entry['sha1']] = '/'.join(
                    [hash_to_hex(dir_id), entry['path'].decode('utf-8')])
                if entry['type'] == 'file':
                    contents.add(entry['sha1'])
                elif entry['type'] == 'dir':
                    directories.add(hash_to_hex(entry['target']))

    # Get all checksums for each content
    result = storage.content_get_metadata(contents)
    contents = []
    for sha1, contents_metadata in result.items():
        for content_metadata in contents_metadata:
            contents.append({
                algo: hash_to_hex(content_metadata[algo])
                for algo in DEFAULT_ALGORITHMS
            })
            path = content_path[sha1]
            cnt = next(storage.content_get([sha1]))
            mimetype, encoding = get_mimetype_and_encoding_for_content(
                cnt['data'])
            _, _, cnt['data'] = _re_encode_content(
                mimetype, encoding, cnt['data'])
            content_display_data = prepare_content_for_display(
                cnt['data'], mimetype, path)
            contents[-1]['path'] = path
            contents[-1]['mimetype'] = mimetype
            contents[-1]['encoding'] = encoding
            contents[-1]['hljs_language'] = content_display_data['language']
            contents[-1]['data'] = content_display_data['content_data']
            _contents[contents[-1]['sha1']] = contents[-1]

    # Create indexer storage instance that will be shared by indexers
    idx_storage = get_indexer_storage('memory', {})

    # Add the empty directory to the test archive
    empty_dir_id = directory_identifier({'entries': []})
    empty_dir_id_bin = hash_to_bytes(empty_dir_id)
    storage.directory_add([{'id': empty_dir_id_bin, 'entries': []}])

    # Return tests data
    return {
        'search': search,
        'storage': storage,
        'idx_storage': idx_storage,
        'origins': _TEST_ORIGINS,
        'contents': contents,
        'directories': list(directories),
        'releases': list(releases),
        'revisions': list(map(hash_to_hex, revisions)),
        'snapshots': list(snapshots),
        'generated_checksums': set(),
    }


def _init_indexers(tests_data):
    # Instantiate content indexers that will be used in tests
    # and force them to use the memory storages
    indexers = {}
    for idx_name, idx_class in (('mimetype_indexer', _MimetypeIndexer),
                                ('license_indexer', _FossologyLicenseIndexer),
                                ('ctags_indexer', _CtagsIndexer)):
        idx = idx_class()
        idx.storage = tests_data['storage']
        idx.objstorage = tests_data['storage'].objstorage
        idx.idx_storage = tests_data['idx_storage']
        idx.register_tools(idx.config['tools'])
        indexers[idx_name] = idx

    return indexers


def get_content(content_sha1):
    return _contents.get(content_sha1)


_tests_data = None
_current_tests_data = None
_indexer_loggers = {}


def get_tests_data(reset=False):
    """
    Initialize tests data and return them in a dict.
    """
    global _tests_data, _current_tests_data
    if _tests_data is None:
        _tests_data = _init_tests_data()
        indexers = _init_indexers(_tests_data)
        for (name, idx) in indexers.items():
            # pytest makes the loggers use a temporary file; and deepcopy
            # requires serializability. So we remove them, and add them
            # back after the copy.
            _indexer_loggers[name] = idx.log
            del idx.log
        _tests_data.update(indexers)
    if reset or _current_tests_data is None:
        _current_tests_data = deepcopy(_tests_data)
        for (name, logger) in _indexer_loggers.items():
            _current_tests_data[name].log = logger
    return _current_tests_data


def override_storages(storage, idx_storage, search):
    """
    Helper function to replace the storages from which archive data
    are fetched.
    """
    swh_config = config.get_config()
    swh_config.update({
        'storage': storage,
        'indexer_storage': idx_storage,
        'search': search,
    })

    service.storage = storage
    service.idx_storage = idx_storage
    service.search = search


# Implement some special endpoints used to provide input tests data
# when executing end to end tests with cypress

_content_code_data_exts = {}  # type: Dict[str, Dict[str, str]]
_content_code_data_filenames = {}  # type: Dict[str, Dict[str, str]]
_content_other_data_exts = {}  # type: Dict[str, Dict[str, str]]


def _init_content_tests_data(data_path, data_dict, ext_key):
    """
    Helper function to read the content of a directory, store it
    into a test archive and add some files metadata (sha1 and/or
    expected programming language) in a dict.

    Args:
        data_path (str): path to a directory relative to the tests
            folder of swh-web
        data_dict (dict): the dict that will store files metadata
        ext_key (bool): whether to use file extensions or filenames
            as dict keys
    """
    test_contents_dir = os.path.join(
        os.path.dirname(__file__), data_path).encode('utf-8')
    directory = Directory.from_disk(path=test_contents_dir, data=True,
                                    save_path=True)
    objects = directory.collect()
    for c in objects['content'].values():
        c['status'] = 'visible'
        sha1 = hash_to_hex(c['sha1'])
        if ext_key:
            key = c['path'].decode('utf-8').split('.')[-1]
            filename = 'test.' + key
        else:
            filename = c['path'].decode('utf-8').split('/')[-1]
            key = filename
        language = get_hljs_language_from_filename(filename)
        data_dict[key] = {'sha1': sha1,
                          'language': language}
        del c['path']
        del c['perms']
    storage = get_tests_data()['storage']
    storage.content_add(objects['content'].values())


def _init_content_code_data_exts():
    """
    Fill a global dictionary which maps source file extension to
    a code content example.
    """
    global _content_code_data_exts
    _init_content_tests_data('resources/contents/code/extensions',
                             _content_code_data_exts, True)


def _init_content_other_data_exts():
    """
    Fill a global dictionary which maps a file extension to
    a content example.
    """
    global _content_other_data_exts
    _init_content_tests_data('resources/contents/other/extensions',
                             _content_other_data_exts, True)


def _init_content_code_data_filenames():
    """
    Fill a global dictionary which maps a filename to
    a content example.
    """
    global _content_code_data_filenames
    _init_content_tests_data('resources/contents/code/filenames',
                             _content_code_data_filenames, False)


if config.get_config()['e2e_tests_mode']:
    _init_content_code_data_exts()
    _init_content_other_data_exts()
    _init_content_code_data_filenames()


@api_view(['GET'])
def get_content_code_data_all_exts(request):
    """
    Endpoint implementation returning a list of all source file
    extensions to test for highlighting using cypress.
    """
    return Response(sorted(_content_code_data_exts.keys()),
                    status=200, content_type='application/json')


@api_view(['GET'])
def get_content_code_data_by_ext(request, ext):
    """
    Endpoint implementation returning metadata of a code content example
    based on the source file extension.
    """
    data = None
    status = 404
    if ext in _content_code_data_exts:
        data = _content_code_data_exts[ext]
        status = 200
    return Response(data, status=status, content_type='application/json')


@api_view(['GET'])
def get_content_other_data_by_ext(request, ext):
    """
    Endpoint implementation returning metadata of a content example
    based on the file extension.
    """
    _init_content_other_data_exts()
    data = None
    status = 404
    if ext in _content_other_data_exts:
        data = _content_other_data_exts[ext]
        status = 200
    return Response(data, status=status, content_type='application/json')


@api_view(['GET'])
def get_content_code_data_all_filenames(request):
    """
    Endpoint implementation returning a list of all source filenames
    to test for highlighting using cypress.
    """
    return Response(sorted(_content_code_data_filenames.keys()),
                    status=200, content_type='application/json')


@api_view(['GET'])
def get_content_code_data_by_filename(request, filename):
    """
    Endpoint implementation returning metadata of a code content example
    based on the source filename.
    """
    data = None
    status = 404
    if filename in _content_code_data_filenames:
        data = _content_code_data_filenames[filename]
        status = 200
    return Response(data, status=status, content_type='application/json')



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from hypothesis import given
import pytest
from requests.utils import parse_header_links

from swh.storage.exc import StorageDBError, StorageAPIError

from swh.web.api.utils import enrich_origin_visit, enrich_origin
from swh.web.common.exc import BadInputExc
from swh.web.common.utils import reverse
from swh.web.common.origin_visits import get_origin_visits
from swh.web.tests.strategies import (
    origin, new_origin, visit_dates, new_snapshots
)


def _scroll_results(api_client, url):
    """Iterates through pages of results, and returns them all."""
    results = []

    while True:
        rv = api_client.get(url)
        assert rv.status_code == 200, rv.data
        assert rv['Content-Type'] == 'application/json'

        results.extend(rv.data)

        if 'Link' in rv:
            for link in parse_header_links(rv['Link']):
                if link['rel'] == 'next':
                    # Found link to next page of results
                    url = link['url']
                    break
            else:
                # No link with 'rel=next'
                break
        else:
            # No Link header
            break

    return results


def test_api_lookup_origin_visits_raise_error(api_client, mocker):
    mock_get_origin_visits = mocker.patch(
        'swh.web.api.views.origin.get_origin_visits')
    err_msg = 'voluntary error to check the bad request middleware.'

    mock_get_origin_visits.side_effect = BadInputExc(err_msg)

    url = reverse('api-1-origin-visits', url_args={'origin_url': 'http://foo'})
    rv = api_client.get(url)

    assert rv.status_code == 400, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'BadInputExc',
        'reason': err_msg
    }


def test_api_lookup_origin_visits_raise_swh_storage_error_db(api_client,
                                                             mocker):
    mock_get_origin_visits = mocker.patch(
        'swh.web.api.views.origin.get_origin_visits')
    err_msg = 'Storage exploded! Will be back online shortly!'

    mock_get_origin_visits.side_effect = StorageDBError(err_msg)

    url = reverse('api-1-origin-visits', url_args={'origin_url': 'http://foo'})
    rv = api_client.get(url)

    assert rv.status_code == 503, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'StorageDBError',
        'reason':
        'An unexpected error occurred in the backend: %s' % err_msg
    }


def test_api_lookup_origin_visits_raise_swh_storage_error_api(api_client,
                                                              mocker):
    mock_get_origin_visits = mocker.patch(
        'swh.web.api.views.origin.get_origin_visits')
    err_msg = 'Storage API dropped dead! Will resurrect asap!'

    mock_get_origin_visits.side_effect = StorageAPIError(err_msg)

    url = reverse(
        'api-1-origin-visits', url_args={'origin_url': 'http://foo'})
    rv = api_client.get(url)

    assert rv.status_code == 503, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'StorageAPIError',
        'reason':
        'An unexpected error occurred in the api backend: %s' % err_msg
    }


@given(new_origin(), visit_dates(3), new_snapshots(3))
def test_api_lookup_origin_visits(api_client, archive_data, new_origin,
                                  visit_dates, new_snapshots):
    archive_data.origin_add_one(new_origin)
    for i, visit_date in enumerate(visit_dates):
        origin_visit = archive_data.origin_visit_add(
            new_origin['url'], visit_date, type='git')
        archive_data.snapshot_add([new_snapshots[i]])
        archive_data.origin_visit_update(
            new_origin['url'], origin_visit['visit'],
            snapshot=new_snapshots[i]['id'])

    all_visits = list(reversed(get_origin_visits(new_origin)))

    for last_visit, expected_visits in (
            (None, all_visits[:2]),
            (all_visits[1]['visit'], all_visits[2:])):

        url = reverse('api-1-origin-visits',
                      url_args={'origin_url': new_origin['url']},
                      query_params={'per_page': 2,
                                    'last_visit': last_visit})

        rv = api_client.get(url)

        assert rv.status_code == 200, rv.data
        assert rv['Content-Type'] == 'application/json'

        for i in range(len(expected_visits)):
            expected_visits[i] = enrich_origin_visit(
                expected_visits[i], with_origin_link=False,
                with_origin_visit_link=True, request=rv.wsgi_request)

        assert rv.data == expected_visits


@given(new_origin(), visit_dates(3), new_snapshots(3))
def test_api_lookup_origin_visits_by_id(api_client, archive_data, new_origin,
                                        visit_dates, new_snapshots):
    archive_data.origin_add_one(new_origin)
    for i, visit_date in enumerate(visit_dates):
        origin_visit = archive_data.origin_visit_add(
            new_origin['url'], visit_date, type='git')
        archive_data.snapshot_add([new_snapshots[i]])
        archive_data.origin_visit_update(
            new_origin['url'], origin_visit['visit'],
            snapshot=new_snapshots[i]['id'])

    all_visits = list(reversed(get_origin_visits(new_origin)))

    for last_visit, expected_visits in (
            (None, all_visits[:2]),
            (all_visits[1]['visit'], all_visits[2:4])):

        url = reverse('api-1-origin-visits',
                      url_args={'origin_url': new_origin['url']},
                      query_params={'per_page': 2,
                                    'last_visit': last_visit})

        rv = api_client.get(url)

        assert rv.status_code == 200, rv.data
        assert rv['Content-Type'] == 'application/json'

        for i in range(len(expected_visits)):
            expected_visits[i] = enrich_origin_visit(
                expected_visits[i], with_origin_link=False,
                with_origin_visit_link=True, request=rv.wsgi_request)

        assert rv.data == expected_visits


@given(new_origin(), visit_dates(3), new_snapshots(3))
def test_api_lookup_origin_visit(api_client, archive_data, new_origin,
                                 visit_dates, new_snapshots):
    archive_data.origin_add_one(new_origin)
    for i, visit_date in enumerate(visit_dates):
        origin_visit = archive_data.origin_visit_add(
            new_origin['url'], visit_date, type='git')
        visit_id = origin_visit['visit']
        archive_data.snapshot_add([new_snapshots[i]])
        archive_data.origin_visit_update(
            new_origin['url'], origin_visit['visit'],
            snapshot=new_snapshots[i]['id'])
        url = reverse('api-1-origin-visit',
                      url_args={'origin_url': new_origin['url'],
                                'visit_id': visit_id})

        rv = api_client.get(url)
        assert rv.status_code == 200, rv.data
        assert rv['Content-Type'] == 'application/json'

        expected_visit = archive_data.origin_visit_get_by(
            new_origin['url'], visit_id)

        expected_visit = enrich_origin_visit(
            expected_visit, with_origin_link=True,
            with_origin_visit_link=False, request=rv.wsgi_request)

        assert rv.data == expected_visit


@given(new_origin())
def test_api_lookup_origin_visit_latest_no_visit(api_client, archive_data,
                                                 new_origin):
    archive_data.origin_add_one(new_origin)

    url = reverse('api-1-origin-visit-latest',
                  url_args={'origin_url': new_origin['url']})

    rv = api_client.get(url)
    assert rv.status_code == 404, rv.data
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'No visit for origin %s found' % new_origin['url']
    }


@given(new_origin(), visit_dates(2), new_snapshots(1))
def test_api_lookup_origin_visit_latest(api_client, archive_data, new_origin,
                                        visit_dates, new_snapshots):
    archive_data.origin_add_one(new_origin)
    visit_dates.sort()
    visit_ids = []
    for i, visit_date in enumerate(visit_dates):
        origin_visit = archive_data.origin_visit_add(
            new_origin['url'], visit_date, type='git')
        visit_ids.append(origin_visit['visit'])

    archive_data.snapshot_add([new_snapshots[0]])
    archive_data.origin_visit_update(
        new_origin['url'], visit_ids[0],
        snapshot=new_snapshots[0]['id'])

    url = reverse('api-1-origin-visit-latest',
                  url_args={'origin_url': new_origin['url']})

    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'

    expected_visit = archive_data.origin_visit_get_by(
        new_origin['url'], visit_ids[1])

    expected_visit = enrich_origin_visit(
            expected_visit, with_origin_link=True,
            with_origin_visit_link=False, request=rv.wsgi_request)

    assert rv.data == expected_visit


@given(new_origin(), visit_dates(2), new_snapshots(1))
def test_api_lookup_origin_visit_latest_with_snapshot(api_client, archive_data,
                                                      new_origin, visit_dates,
                                                      new_snapshots):
    archive_data.origin_add_one(new_origin)
    visit_dates.sort()
    visit_ids = []
    for i, visit_date in enumerate(visit_dates):
        origin_visit = archive_data.origin_visit_add(
            new_origin['url'], visit_date, type='git')
        visit_ids.append(origin_visit['visit'])

    archive_data.snapshot_add([new_snapshots[0]])
    archive_data.origin_visit_update(
        new_origin['url'], visit_ids[0],
        snapshot=new_snapshots[0]['id'])

    url = reverse('api-1-origin-visit-latest',
                  url_args={'origin_url': new_origin['url']},
                  query_params={'require_snapshot': True})

    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'

    expected_visit = archive_data.origin_visit_get_by(
        new_origin['url'], visit_ids[0])

    expected_visit = enrich_origin_visit(
            expected_visit, with_origin_link=True,
            with_origin_visit_link=False, request=rv.wsgi_request)

    assert rv.data == expected_visit


@given(origin())
def test_api_lookup_origin_visit_not_found(api_client, origin):

    all_visits = list(reversed(get_origin_visits(origin)))

    max_visit_id = max([v['visit'] for v in all_visits])

    url = reverse('api-1-origin-visit',
                  url_args={'origin_url': origin['url'],
                            'visit_id': max_visit_id + 1})

    rv = api_client.get(url)

    assert rv.status_code == 404, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'Origin %s or its visit with id %s not found!' %
        (origin['url'], max_visit_id+1)
    }


def test_api_origins(api_client, archive_data):
    origins = list(archive_data.origin_get_range(0, 10000))
    origin_urls = {origin['url'] for origin in origins}

    # Get only one
    url = reverse('api-1-origins',
                  query_params={'origin_count': 1})
    rv = api_client.get(url)
    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert len(rv.data) == 1
    assert {origin['url'] for origin in rv.data} <= origin_urls

    # Get all
    url = reverse('api-1-origins',
                  query_params={'origin_count': len(origins)})
    rv = api_client.get(url)
    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert len(rv.data) == len(origins)
    assert {origin['url'] for origin in rv.data} == origin_urls

    # Get "all + 10"
    url = reverse('api-1-origins',
                  query_params={'origin_count': len(origins)+10})
    rv = api_client.get(url)
    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert len(rv.data) == len(origins)
    assert {origin['url'] for origin in rv.data} == origin_urls


@pytest.mark.parametrize('origin_count', [1, 2, 10, 100])
def test_api_origins_scroll(api_client, archive_data, origin_count):
    origins = list(archive_data.origin_get_range(0, 10000))
    origin_urls = {origin['url'] for origin in origins}

    url = reverse('api-1-origins',
                  query_params={'origin_count': origin_count})

    results = _scroll_results(api_client, url)

    assert len(results) == len(origins)
    assert {origin['url'] for origin in results} == origin_urls


@given(origin())
def test_api_origin_by_url(api_client, archive_data, origin):
    url = reverse('api-1-origin',
                  url_args={'origin_url': origin['url']})
    rv = api_client.get(url)

    expected_origin = archive_data.origin_get(origin)

    expected_origin = enrich_origin(expected_origin, rv.wsgi_request)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == expected_origin


@given(new_origin())
def test_api_origin_not_found(api_client, new_origin):

    url = reverse('api-1-origin',
                  url_args={'origin_url': new_origin['url']})
    rv = api_client.get(url)

    assert rv.status_code == 404, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'Origin with url %s not found!' % new_origin['url']
    }


@pytest.mark.parametrize('backend', ['swh-search', 'swh-storage'])
def test_api_origin_search(api_client, mocker, backend):
    if backend != 'swh-search':
        # equivalent to not configuring search in the config
        mocker.patch('swh.web.common.service.search', None)

    expected_origins = {
        'https://github.com/wcoder/highlightjs-line-numbers.js',
        'https://github.com/memononen/libtess2',
    }

    # Search for 'github.com', get only one
    url = reverse('api-1-origin-search',
                  url_args={'url_pattern': 'github.com'},
                  query_params={'limit': 1})
    rv = api_client.get(url)
    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert len(rv.data) == 1
    assert {origin['url'] for origin in rv.data} <= expected_origins

    # Search for 'github.com', get all
    url = reverse('api-1-origin-search',
                  url_args={'url_pattern': 'github.com'},
                  query_params={'limit': 2})
    rv = api_client.get(url)
    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert {origin['url'] for origin in rv.data} == expected_origins

    # Search for 'github.com', get more than available
    url = reverse('api-1-origin-search',
                  url_args={'url_pattern': 'github.com'},
                  query_params={'limit': 10})
    rv = api_client.get(url)
    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert {origin['url'] for origin in rv.data} == expected_origins


@pytest.mark.parametrize('backend', ['swh-search', 'swh-storage'])
def test_api_origin_search_words(api_client, mocker, backend):
    if backend != 'swh-search':
        # equivalent to not configuring search in the config
        mocker.patch('swh.web.common.service.search', None)

    expected_origins = {
        'https://github.com/wcoder/highlightjs-line-numbers.js',
        'https://github.com/memononen/libtess2',
    }

    url = reverse('api-1-origin-search',
                  url_args={'url_pattern': 'github com'},
                  query_params={'limit': 2})
    rv = api_client.get(url)
    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert {origin['url'] for origin in rv.data} == expected_origins

    url = reverse('api-1-origin-search',
                  url_args={'url_pattern': 'com github'},
                  query_params={'limit': 2})
    rv = api_client.get(url)
    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert {origin['url'] for origin in rv.data} == expected_origins

    url = reverse('api-1-origin-search',
                  url_args={'url_pattern': 'memononen libtess2'},
                  query_params={'limit': 2})
    rv = api_client.get(url)
    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert len(rv.data) == 1
    assert {origin['url'] for origin in rv.data} \
        == {'https://github.com/memononen/libtess2'}

    url = reverse('api-1-origin-search',
                  url_args={'url_pattern': 'libtess2 memononen'},
                  query_params={'limit': 2})
    rv = api_client.get(url)
    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert len(rv.data) == 1
    assert {origin['url'] for origin in rv.data} \
        == {'https://github.com/memononen/libtess2'}


@pytest.mark.parametrize('backend', ['swh-search', 'swh-storage'])
@pytest.mark.parametrize('limit', [1, 2, 3, 10])
def test_api_origin_search_scroll(
        api_client, archive_data, mocker, limit, backend):

    if backend != 'swh-search':
        # equivalent to not configuring search in the config
        mocker.patch('swh.web.common.service.search', None)

    expected_origins = {
        'https://github.com/wcoder/highlightjs-line-numbers.js',
        'https://github.com/memononen/libtess2',
    }

    url = reverse('api-1-origin-search',
                  url_args={'url_pattern': 'github.com'},
                  query_params={'limit': limit})

    results = _scroll_results(api_client, url)

    assert {origin['url'] for origin in results} == expected_origins


@pytest.mark.parametrize('backend', ['swh-search', 'swh-storage'])
def test_api_origin_search_limit(
        api_client, archive_data, tests_data, mocker, backend):
    if backend == 'swh-search':
        tests_data['search'].origin_update([
            {'url': 'http://foobar/{}'.format(i)}
            for i in range(2000)
        ])
    else:
        # equivalent to not configuring search in the config
        mocker.patch('swh.web.common.service.search', None)

        archive_data.origin_add([
            {'url': 'http://foobar/{}'.format(i)}
            for i in range(2000)
        ])

    url = reverse('api-1-origin-search',
                  url_args={'url_pattern': 'foobar'},
                  query_params={'limit': 1050})
    rv = api_client.get(url)
    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert len(rv.data) == 1000


@given(origin())
def test_api_origin_metadata_search(api_client, mocker, origin):
    mock_idx_storage = mocker.patch('swh.web.common.service.idx_storage')
    oimsft = mock_idx_storage.origin_intrinsic_metadata_search_fulltext
    oimsft.side_effect = lambda conjunction, limit: [{
        'from_revision': (
            b'p&\xb7\xc1\xa2\xafVR\x1e\x95\x1c\x01\xed '
            b'\xf2U\xfa\x05B8'),
        'metadata': {'author': 'Jane Doe'},
        'id': origin['url'],
        'tool': {
            'configuration': {
                'context': ['NpmMapping', 'CodemetaMapping'],
                'type': 'local'
            },
            'id': 3,
            'name': 'swh-metadata-detector',
            'version': '0.0.1'
        }
    }]

    url = reverse('api-1-origin-metadata-search',
                  query_params={'fulltext': 'Jane Doe'})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.content
    assert rv['Content-Type'] == 'application/json'
    expected_data = [{
        'url': origin['url'],
        'metadata': {
            'metadata': {'author': 'Jane Doe'},
            'from_revision': (
                '7026b7c1a2af56521e951c01ed20f255fa054238'),
            'tool': {
                'configuration': {
                    'context': ['NpmMapping', 'CodemetaMapping'],
                    'type': 'local'
                },
                'id': 3,
                'name': 'swh-metadata-detector',
                'version': '0.0.1',
            }
        }
    }]

    assert rv.data == expected_data
    oimsft.assert_called_with(conjunction=['Jane Doe'], limit=70)


@given(origin())
def test_api_origin_metadata_search_limit(api_client, mocker, origin):
    mock_idx_storage = mocker.patch('swh.web.common.service.idx_storage')
    oimsft = mock_idx_storage.origin_intrinsic_metadata_search_fulltext

    oimsft.side_effect = lambda conjunction, limit: [{
        'from_revision': (
            b'p&\xb7\xc1\xa2\xafVR\x1e\x95\x1c\x01\xed '
            b'\xf2U\xfa\x05B8'),
        'metadata': {'author': 'Jane Doe'},
        'id': origin['url'],
        'tool': {
            'configuration': {
                'context': ['NpmMapping', 'CodemetaMapping'],
                'type': 'local'
            },
            'id': 3,
            'name': 'swh-metadata-detector',
            'version': '0.0.1'
        }
    }]

    url = reverse('api-1-origin-metadata-search',
                  query_params={'fulltext': 'Jane Doe'})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.content
    assert rv['Content-Type'] == 'application/json'
    assert len(rv.data) == 1
    oimsft.assert_called_with(conjunction=['Jane Doe'], limit=70)

    url = reverse('api-1-origin-metadata-search',
                  query_params={'fulltext': 'Jane Doe',
                                'limit': 10})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.content
    assert rv['Content-Type'] == 'application/json'
    assert len(rv.data) == 1
    oimsft.assert_called_with(conjunction=['Jane Doe'], limit=10)

    url = reverse('api-1-origin-metadata-search',
                  query_params={'fulltext': 'Jane Doe',
                                'limit': 987})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.content
    assert rv['Content-Type'] == 'application/json'
    assert len(rv.data) == 1
    oimsft.assert_called_with(conjunction=['Jane Doe'], limit=100)


@given(origin())
def test_api_origin_intrinsic_metadata(api_client, mocker, origin):
    mock_idx_storage = mocker.patch('swh.web.common.service.idx_storage')
    oimg = mock_idx_storage.origin_intrinsic_metadata_get
    oimg.side_effect = lambda origin_urls: [{
        'from_revision': (
            b'p&\xb7\xc1\xa2\xafVR\x1e\x95\x1c\x01\xed '
            b'\xf2U\xfa\x05B8'),
        'metadata': {'author': 'Jane Doe'},
        'id': origin['url'],
        'tool': {
            'configuration': {
                'context': ['NpmMapping', 'CodemetaMapping'],
                'type': 'local'
            },
            'id': 3,
            'name': 'swh-metadata-detector',
            'version': '0.0.1'
        }
    }]

    url = reverse('api-origin-intrinsic-metadata',
                  url_args={'origin_url': origin['url']})
    rv = api_client.get(url)

    oimg.assert_called_once_with([origin['url']])
    assert rv.status_code == 200, rv.content
    assert rv['Content-Type'] == 'application/json'
    expected_data = {'author': 'Jane Doe'}
    assert rv.data == expected_data


def test_api_origin_metadata_search_invalid(api_client, mocker):
    mock_idx_storage = mocker.patch('swh.web.common.service.idx_storage')
    url = reverse('api-1-origin-metadata-search')
    rv = api_client.get(url)

    assert rv.status_code == 400, rv.content
    mock_idx_storage.assert_not_called()



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.web.common import service
from swh.web.api import utils
from swh.web.api.apidoc import api_doc, format_docstring
from swh.web.api.apiurls import api_route
from swh.web.api.views.utils import api_lookup


@api_route(r'/directory/(?P<sha1_git>[0-9a-f]+)/', 'api-1-directory',
           checksum_args=['sha1_git'])
@api_route(r'/directory/(?P<sha1_git>[0-9a-f]+)/(?P<path>.+)/',
           'api-1-directory',
           checksum_args=['sha1_git'])
@api_doc('/directory/')
@format_docstring()
def api_directory(request, sha1_git, path=None):
    """
    .. http:get:: /api/1/directory/(sha1_git)/[(path)/]

        Get information about directory objects.
        Directories are identified by **sha1** checksums, compatible with Git directory identifiers.
        See :func:`swh.model.identifiers.directory_identifier` in our data model module for details
        about how they are computed.

        When given only a directory identifier, this endpoint returns information about the directory itself,
        returning its content (usually a list of directory entries). When given a directory identifier and a
        path, this endpoint returns information about the directory entry pointed by the relative path,
        starting path resolution from the given directory.

        :param string sha1_git: hexadecimal representation of the directory **sha1_git** identifier
        :param string path: optional parameter to get information about the directory entry
            pointed by that relative path

        {common_headers}

        :>jsonarr object checksums: object holding the computed checksum values for a directory entry
            (only for file entries)
        :>jsonarr string dir_id: **sha1_git** identifier of the requested directory
        :>jsonarr number length: length of a directory entry in bytes (only for file entries)
            for getting information about the content MIME type
        :>jsonarr string name: the directory entry name
        :>jsonarr number perms: permissions for the directory entry
        :>jsonarr string target: **sha1_git** identifier of the directory entry
        :>jsonarr string target_url: link to :http:get:`/api/1/content/[(hash_type):](hash)/`
            or :http:get:`/api/1/directory/(sha1_git)/[(path)/]` depending on the directory entry type
        :>jsonarr string type: the type of the directory entry, can be either ``dir``, ``file`` or ``rev``

        **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`, :http:method:`options`

        :statuscode 200: no error
        :statuscode 400: an invalid **hash_type** or **hash** has been provided
        :statuscode 404: requested directory can not be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`directory/977fc4b98c0e85816348cebd3b12026407c368b6/`
    """ # noqa
    if path:
        error_msg_path = ('Entry with path %s relative to directory '
                          'with sha1_git %s not found.') % (path, sha1_git)
        return api_lookup(
            service.lookup_directory_with_path, sha1_git, path,
            notfound_msg=error_msg_path,
            enrich_fn=utils.enrich_directory,
            request=request)
    else:
        error_msg_nopath = 'Directory with sha1_git %s not found.' % sha1_git
        return api_lookup(
            service.lookup_directory, sha1_git,
            notfound_msg=error_msg_nopath,
            enrich_fn=utils.enrich_directory,
            request=request)



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from urllib.parse import unquote

import pytest

from django.contrib.auth import get_user_model

from swh.web.common.models import (
    SaveAuthorizedOrigin, SaveUnauthorizedOrigin, SaveOriginRequest
)
from swh.web.common.origin_save import can_save_origin
from swh.web.common.models import (
    SAVE_REQUEST_PENDING, SAVE_REQUEST_ACCEPTED,
    SAVE_REQUEST_REJECTED, SAVE_TASK_NOT_YET_SCHEDULED
)
from swh.web.common.utils import reverse

_user_name = 'swh-web-admin'
_user_mail = 'admin@swh-web.org'
_user_password = '..34~pounds~BEAUTY~march~63..'

_authorized_origin_url = 'https://scm.ourproject.org/anonscm/'
_unauthorized_origin_url = 'https://www.softwareheritage.org/'


pytestmark = pytest.mark.django_db


@pytest.fixture(autouse=True)
def populated_db():
    User = get_user_model()
    user = User.objects.create_user(_user_name, _user_mail, _user_password)
    user.is_staff = True
    user.save()
    SaveAuthorizedOrigin.objects.create(url=_authorized_origin_url)
    SaveUnauthorizedOrigin.objects.create(url=_unauthorized_origin_url)


def check_not_login(client, url):
    login_url = reverse('login', query_params={'next': url})
    response = client.post(url)
    assert response.status_code == 302
    assert unquote(response.url) == login_url


def test_add_authorized_origin_url(client):
    authorized_url = 'https://scm.adullact.net/anonscm/'
    assert can_save_origin(authorized_url) == SAVE_REQUEST_PENDING

    url = reverse('admin-origin-save-add-authorized-url',
                  url_args={'origin_url': authorized_url})

    check_not_login(client, url)

    assert can_save_origin(authorized_url) == SAVE_REQUEST_PENDING

    client.login(username=_user_name, password=_user_password)
    response = client.post(url)
    assert response.status_code == 200
    assert can_save_origin(authorized_url) == SAVE_REQUEST_ACCEPTED


def test_remove_authorized_origin_url(client):
    assert can_save_origin(_authorized_origin_url) == SAVE_REQUEST_ACCEPTED

    url = reverse('admin-origin-save-remove-authorized-url',
                  url_args={'origin_url': _authorized_origin_url})

    check_not_login(client, url)

    assert can_save_origin(_authorized_origin_url) == SAVE_REQUEST_ACCEPTED

    client.login(username=_user_name, password=_user_password)
    response = client.post(url)
    assert response.status_code == 200
    assert can_save_origin(_authorized_origin_url) == SAVE_REQUEST_PENDING


def test_add_unauthorized_origin_url(client):
    unauthorized_url = 'https://www.yahoo./'
    assert can_save_origin(unauthorized_url) == SAVE_REQUEST_PENDING

    url = reverse('admin-origin-save-add-unauthorized-url',
                  url_args={'origin_url': unauthorized_url})

    check_not_login(client, url)

    assert can_save_origin(unauthorized_url) == SAVE_REQUEST_PENDING

    client.login(username=_user_name, password=_user_password)
    response = client.post(url)
    assert response.status_code == 200
    assert can_save_origin(unauthorized_url) == SAVE_REQUEST_REJECTED


def test_remove_unauthorized_origin_url(client):
    assert can_save_origin(_unauthorized_origin_url) == SAVE_REQUEST_REJECTED

    url = reverse('admin-origin-save-remove-unauthorized-url',
                  url_args={'origin_url': _unauthorized_origin_url})

    check_not_login(client, url)

    assert can_save_origin(_unauthorized_origin_url) == SAVE_REQUEST_REJECTED

    client.login(username=_user_name, password=_user_password)
    response = client.post(url)
    assert response.status_code == 200
    assert can_save_origin(_unauthorized_origin_url) == SAVE_REQUEST_PENDING


def test_accept_pending_save_request(client, mocker):
    mock_scheduler = mocker.patch('swh.web.common.origin_save.scheduler')
    visit_type = 'git'
    origin_url = 'https://v2.pikacode.com/bthate/botlib.git'
    save_request_url = reverse('api-1-save-origin',
                               url_args={'visit_type': visit_type,
                                         'origin_url': origin_url})
    response = client.post(save_request_url, data={},
                           content_type='application/x-www-form-urlencoded')
    assert response.status_code == 200
    assert response.data['save_request_status'] == SAVE_REQUEST_PENDING

    accept_request_url = reverse('admin-origin-save-request-accept',
                                 url_args={'visit_type': visit_type,
                                           'origin_url': origin_url})

    check_not_login(client, accept_request_url)

    tasks_data = [
        {
            'priority': 'high',
            'policy': 'oneshot',
            'type': 'load-git',
            'arguments': {
                'kwargs': {
                    'repo_url': origin_url
                },
                'args': []
            },
            'status': 'next_run_not_scheduled',
            'id': 1,
            }
    ]

    mock_scheduler.create_tasks.return_value = tasks_data
    mock_scheduler.get_tasks.return_value = tasks_data

    client.login(username=_user_name, password=_user_password)
    response = client.post(accept_request_url)
    assert response.status_code == 200

    response = client.get(save_request_url)
    assert response.status_code == 200
    assert response.data[0]['save_request_status'] == SAVE_REQUEST_ACCEPTED
    assert response.data[0]['save_task_status'] == SAVE_TASK_NOT_YET_SCHEDULED


def test_reject_pending_save_request(client, mocker):
    mock_scheduler = mocker.patch('swh.web.common.origin_save.scheduler')
    visit_type = 'git'
    origin_url = 'https://wikipedia.com'
    save_request_url = reverse('api-1-save-origin',
                               url_args={'visit_type': visit_type,
                                         'origin_url': origin_url})
    response = client.post(save_request_url, data={},
                           content_type='application/x-www-form-urlencoded')
    assert response.status_code == 200
    assert response.data['save_request_status'] == SAVE_REQUEST_PENDING

    reject_request_url = reverse('admin-origin-save-request-reject',
                                 url_args={'visit_type': visit_type,
                                           'origin_url': origin_url})

    check_not_login(client, reject_request_url)

    client.login(username=_user_name, password=_user_password)
    response = client.post(reject_request_url)
    assert response.status_code == 200

    tasks_data = [
        {
            'priority': 'high',
            'policy': 'oneshot',
            'type': 'load-git',
            'arguments': {
                'kwargs': {
                    'repo_url': origin_url
                },
                'args': []
            },
            'status': 'next_run_not_scheduled',
            'id': 1,
        }
    ]

    mock_scheduler.create_tasks.return_value = tasks_data
    mock_scheduler.get_tasks.return_value = tasks_data

    response = client.get(save_request_url)
    assert response.status_code == 200
    assert response.data[0]['save_request_status'] == SAVE_REQUEST_REJECTED


def test_remove_save_request(client):
    sor = SaveOriginRequest.objects.create(visit_type='git',
                                            origin_url='https://wikipedia.com', # noqa
                                            status=SAVE_REQUEST_PENDING)
    assert SaveOriginRequest.objects.count() == 1

    remove_request_url = reverse('admin-origin-save-request-remove',
                                 url_args={'sor_id': sor.id})

    check_not_login(client, remove_request_url)

    client.login(username=_user_name, password=_user_password)
    response = client.post(remove_request_url)
    assert response.status_code == 200
    assert SaveOriginRequest.objects.count() == 0



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random

from hypothesis import given

from swh.web.common.utils import reverse, get_swh_persistent_id
from swh.web.common.utils import gen_path_info
from swh.web.tests.django_asserts import assert_contains, assert_template_used
from swh.web.tests.strategies import (
    directory, directory_with_subdirs, invalid_sha1,
    unknown_directory
)


@given(directory())
def test_root_directory_view(client, archive_data, directory):
    _directory_view(client, directory, archive_data.directory_ls(directory))


@given(directory_with_subdirs())
def test_sub_directory_view(client, archive_data, directory):
    dir_content = archive_data.directory_ls(directory)
    subdir = random.choice([e for e in dir_content if e['type'] == 'dir'])
    subdir_content = archive_data.directory_ls(subdir['target'])
    _directory_view(client, directory, subdir_content, subdir['name'])


@given(invalid_sha1(), unknown_directory())
def test_directory_request_errors(client, invalid_sha1, unknown_directory):
    dir_url = reverse('browse-directory',
                      url_args={'sha1_git': invalid_sha1})

    resp = client.get(dir_url)
    assert resp.status_code == 400
    assert_template_used(resp, 'error.html')

    dir_url = reverse('browse-directory',
                      url_args={'sha1_git': unknown_directory})

    resp = client.get(dir_url)
    assert resp.status_code == 404
    assert_template_used(resp, 'error.html')


@given(directory())
def test_directory_uppercase(client, directory):
    url = reverse('browse-directory-uppercase-checksum',
                  url_args={'sha1_git': directory.upper()})

    resp = client.get(url)
    assert resp.status_code == 302

    redirect_url = reverse('browse-directory',
                           url_args={'sha1_git': directory})

    assert resp['location'] == redirect_url


def _directory_view(client, root_directory_sha1, directory_entries,
                    path=None):
    dirs = [e for e in directory_entries if e['type'] in ('dir', 'rev')]
    files = [e for e in directory_entries if e['type'] == 'file']

    url_args = {'sha1_git': root_directory_sha1}
    if path:
        url_args['path'] = path

    url = reverse('browse-directory',
                  url_args=url_args)

    root_dir_url = reverse('browse-directory',
                           url_args={'sha1_git': root_directory_sha1})

    resp = client.get(url)

    assert resp.status_code == 200
    assert_template_used(resp, 'browse/directory.html')
    assert_contains(resp, '<a href="' + root_dir_url + '">' +
                    root_directory_sha1[:7] + '</a>')
    assert_contains(resp, '<td class="swh-directory">',
                    count=len(dirs))
    assert_contains(resp, '<td class="swh-content">',
                    count=len(files))

    for d in dirs:
        if d['type'] == 'rev':
            dir_url = reverse('browse-revision',
                              url_args={'sha1_git': d['target']})
        else:
            dir_path = d['name']
            if path:
                dir_path = "%s/%s" % (path, d['name'])
            dir_url = reverse('browse-directory',
                              url_args={'sha1_git': root_directory_sha1,
                                        'path': dir_path})
        assert_contains(resp, dir_url)

    for f in files:
        file_path = "%s/%s" % (root_directory_sha1, f['name'])
        if path:
            file_path = "%s/%s/%s" % (root_directory_sha1, path, f['name'])
        query_string = 'sha1_git:' + f['target']
        file_url = reverse('browse-content',
                           url_args={'query_string': query_string},
                           query_params={'path': file_path})
        assert_contains(resp, file_url)

    path_info = gen_path_info(path)

    assert_contains(resp, '<li class="swh-path">',
                    count=len(path_info)+1)
    assert_contains(resp, '<a href="%s">%s</a>' %
                    (root_dir_url, root_directory_sha1[:7]))

    for p in path_info:
        dir_url = reverse('browse-directory',
                          url_args={'sha1_git': root_directory_sha1,
                                    'path': p['path']})
        assert_contains(resp, '<a href="%s">%s</a>' %
                        (dir_url, p['name']))

    assert_contains(resp, 'vault-cook-directory')

    swh_dir_id = get_swh_persistent_id(
        'directory', directory_entries[0]['dir_id'])
    swh_dir_id_url = reverse('browse-swh-id',
                             url_args={'swh_id': swh_dir_id})
    assert_contains(resp, swh_dir_id)
    assert_contains(resp, swh_dir_id_url)



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import functools

from typing import Dict

from rest_framework.decorators import api_view

from swh.web.common.urlsindex import UrlsIndex
from swh.web.common import throttling


class APIUrls(UrlsIndex):
    """
    Class to manage API documentation URLs.

    - Indexes all routes documented using apidoc's decorators.
    - Tracks endpoint/request processing method relationships for use in
      generating related urls in API documentation

    """
    _apidoc_routes = {}  # type: Dict[str, Dict[str, str]]
    scope = 'api'

    @classmethod
    def get_app_endpoints(cls):
        return cls._apidoc_routes

    @classmethod
    def add_route(cls, route, docstring, **kwargs):
        """
        Add a route to the self-documenting API reference
        """
        route_view_name = 'api-1-%s' % route[1:-1].replace('/', '-')
        if route not in cls._apidoc_routes:
            d = {'docstring': docstring,
                 'route_view_name': route_view_name}
            for k, v in kwargs.items():
                d[k] = v
            cls._apidoc_routes[route] = d


def api_route(url_pattern=None, view_name=None,
              methods=['GET', 'HEAD', 'OPTIONS'],
              throttle_scope='swh_api',
              api_version='1',
              checksum_args=None):
    """
    Decorator to ease the registration of an API endpoint
    using the Django REST Framework.

    Args:
        url_pattern: the url pattern used by DRF to identify the API route
        view_name: the name of the API view associated to the route used to
           reverse the url
        methods: array of HTTP methods supported by the API route

    """

    url_pattern = '^' + api_version + url_pattern + '$'

    def decorator(f):
        # create a DRF view from the wrapped function
        @api_view(methods)
        @throttling.throttle_scope(throttle_scope)
        @functools.wraps(f)
        def api_view_f(*args, **kwargs):
            return f(*args, **kwargs)
        # small hacks for correctly generating API endpoints index doc
        api_view_f.__name__ = f.__name__
        api_view_f.http_method_names = methods

        # register the route and its view in the endpoints index
        APIUrls.add_url_pattern(url_pattern, api_view_f,
                                view_name)

        if checksum_args:
            APIUrls.add_redirect_for_checksum_args(view_name,
                                                   [url_pattern],
                                                   checksum_args)
        return f

    return decorator



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from django.shortcuts import redirect

from swh.web.browse.browseurls import browse_route
from swh.web.common.utils import reverse

from .utils.snapshot_context import (
    browse_snapshot_directory, browse_snapshot_content,
    browse_snapshot_log, browse_snapshot_branches,
    browse_snapshot_releases
)


@browse_route(r'snapshot/(?P<snapshot_id>[0-9a-f]+)/',
              view_name='browse-snapshot',
              checksum_args=['snapshot_id'])
def snapshot_browse(request, snapshot_id):
    """Django view for browsing the content of a snapshot.

    The url that points to it is :http:get:`/browse/snapshot/(snapshot_id)/`
    """
    browse_snapshot_url = reverse('browse-snapshot-directory',
                                  url_args={'snapshot_id': snapshot_id},
                                  query_params=request.GET)
    return redirect(browse_snapshot_url)


@browse_route(r'snapshot/(?P<snapshot_id>[0-9a-f]+)/directory/',
              r'snapshot/(?P<snapshot_id>[0-9a-f]+)/directory/(?P<path>.+)/',
              view_name='browse-snapshot-directory',
              checksum_args=['snapshot_id'])
def snapshot_directory_browse(request, snapshot_id, path=None):
    """Django view for browsing the content of a directory collected
    in a snapshot.

    The url that points to it is
    :http:get:`/browse/snapshot/(snapshot_id)/directory/[(path)/]`
    """
    origin_url = request.GET.get('origin_url', None)
    if not origin_url:
        origin_url = request.GET.get('origin', None)
    return browse_snapshot_directory(request, snapshot_id=snapshot_id,
                                     path=path, origin_url=origin_url)


@browse_route(r'snapshot/(?P<snapshot_id>[0-9a-f]+)/content/(?P<path>.+)/',
              view_name='browse-snapshot-content',
              checksum_args=['snapshot_id'])
def snapshot_content_browse(request, snapshot_id, path):
    """Django view that produces an HTML display of a content
    collected in a snapshot.

    The url that points to it is
    :http:get:`/browse/snapshot/(snapshot_id)/content/(path)/`
    """
    language = request.GET.get('language', None)
    return browse_snapshot_content(request, snapshot_id=snapshot_id, path=path,
                                   selected_language=language)


@browse_route(r'snapshot/(?P<snapshot_id>[0-9a-f]+)/log/',
              view_name='browse-snapshot-log',
              checksum_args=['snapshot_id'])
def snapshot_log_browse(request, snapshot_id):
    """Django view that produces an HTML display of revisions history (aka
    the commit log) collected in a snapshot.

    The url that points to it is
    :http:get:`/browse/snapshot/(snapshot_id)/log/`
    """
    return browse_snapshot_log(request, snapshot_id=snapshot_id)


@browse_route(r'snapshot/(?P<snapshot_id>[0-9a-f]+)/branches/',
              view_name='browse-snapshot-branches',
              checksum_args=['snapshot_id'])
def snapshot_branches_browse(request, snapshot_id):
    """Django view that produces an HTML display of the list of releases
    collected in a snapshot.

    The url that points to it is
    :http:get:`/browse/snapshot/(snapshot_id)/branches/`
    """
    return browse_snapshot_branches(request, snapshot_id=snapshot_id)


@browse_route(r'snapshot/(?P<snapshot_id>[0-9a-f]+)/releases/',
              view_name='browse-snapshot-releases',
              checksum_args=['snapshot_id'])
def snapshot_releases_browse(request, snapshot_id):
    """Django view that produces an HTML display of the list of releases
    collected in a snapshot.

    The url that points to it is
    :http:get:`/browse/snapshot/(snapshot_id)/releases/`
    """
    return browse_snapshot_releases(request, snapshot_id=snapshot_id)



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.model import hashutil
from swh.web.common.utils import reverse

TEST_OBJ_ID = 'd4905454cc154b492bd6afed48694ae3c579345e'

OBJECT_TYPES = ('directory', 'revision_gitfast')


def test_api_vault_cook(api_client, mocker):
    mock_service = mocker.patch('swh.web.api.views.vault.service')

    for obj_type in OBJECT_TYPES:

        fetch_url = reverse(f'api-1-vault-fetch-{obj_type}',
                            url_args={f'{obj_type[:3]}_id': TEST_OBJ_ID})
        stub_cook = {
            'fetch_url': fetch_url,
            'obj_id': TEST_OBJ_ID,
            'obj_type': obj_type,
            'progress_message': None,
            'status': 'done',
            'task_uuid': 'de75c902-5ee5-4739-996e-448376a93eff',
        }
        stub_fetch = b'content'

        mock_service.vault_cook.return_value = stub_cook
        mock_service.vault_fetch.return_value = stub_fetch

        url = reverse(f'api-1-vault-cook-{obj_type}',
                      url_args={f'{obj_type[:3]}_id': TEST_OBJ_ID})

        rv = api_client.post(url, {'email': 'test@test.mail'})

        assert rv.status_code == 200, rv.data
        assert rv['Content-Type'] == 'application/json'

        stub_cook['fetch_url'] = rv.wsgi_request.build_absolute_uri(
            stub_cook['fetch_url'])

        assert rv.data == stub_cook
        mock_service.vault_cook.assert_called_with(
            obj_type,
            hashutil.hash_to_bytes(TEST_OBJ_ID),
            'test@test.mail')

        rv = api_client.get(fetch_url)

        assert rv.status_code == 200
        assert rv['Content-Type'] == 'application/gzip'
        assert rv.content == stub_fetch
        mock_service.vault_fetch.assert_called_with(
            obj_type, hashutil.hash_to_bytes(TEST_OBJ_ID))


def test_api_vault_cook_uppercase_hash(api_client):

    for obj_type in OBJECT_TYPES:

        url = reverse(f'api-1-vault-cook-{obj_type}-uppercase-checksum',
                      url_args={f'{obj_type[:3]}_id': TEST_OBJ_ID.upper()})
        rv = api_client.post(url, {'email': 'test@test.mail'})

        assert rv.status_code == 302

        redirect_url = reverse(f'api-1-vault-cook-{obj_type}',
                               url_args={f'{obj_type[:3]}_id': TEST_OBJ_ID})

        assert rv['location'] == redirect_url

        fetch_url = reverse(
            f'api-1-vault-fetch-{obj_type}-uppercase-checksum',
            url_args={f'{obj_type[:3]}_id': TEST_OBJ_ID.upper()})

        rv = api_client.get(fetch_url)

        assert rv.status_code == 302

        redirect_url = reverse(f'api-1-vault-fetch-{obj_type}',
                               url_args={f'{obj_type[:3]}_id': TEST_OBJ_ID})

        assert rv['location'] == redirect_url


def test_api_vault_cook_notfound(api_client, mocker):
    mock_service = mocker.patch('swh.web.api.views.vault.service')
    mock_service.vault_cook.return_value = None
    mock_service.vault_fetch.return_value = None

    for obj_type in OBJECT_TYPES:
        url = reverse(f'api-1-vault-cook-{obj_type}',
                      url_args={f'{obj_type[:3]}_id': TEST_OBJ_ID})
        rv = api_client.post(url)

        assert rv.status_code == 404, rv.data
        assert rv['Content-Type'] == 'application/json'

        assert rv.data['exception'] == 'NotFoundExc'
        mock_service.vault_cook.assert_called_with(
            obj_type, hashutil.hash_to_bytes(TEST_OBJ_ID), None)

        fetch_url = reverse(f'api-1-vault-fetch-{obj_type}',
                            url_args={f'{obj_type[:3]}_id': TEST_OBJ_ID})

        rv = api_client.get(fetch_url)

        assert rv.status_code == 404, rv.data
        assert rv['Content-Type'] == 'application/json'
        assert rv.data['exception'] == 'NotFoundExc'
        mock_service.vault_fetch.assert_called_with(
            obj_type, hashutil.hash_to_bytes(TEST_OBJ_ID))



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from rest_framework.response import Response

from swh.storage.exc import StorageDBError, StorageAPIError

from swh.web.api.apidoc import api_doc, _parse_httpdomain_doc
from swh.web.api.apiurls import api_route
from swh.web.common.exc import BadInputExc, ForbiddenExc, NotFoundExc
from swh.web.tests.django_asserts import assert_template_used


httpdomain_doc = """
.. http:get:: /api/1/revision/(sha1_git)/

    Get information about a revision in the archive.
    Revisions are identified by **sha1** checksums, compatible with Git commit
    identifiers.
    See :func:`swh.model.identifiers.revision_identifier` in our data model
    module for details about how they are computed.

    :param string sha1_git: hexadecimal representation of the revision
        **sha1_git** identifier

    :reqheader Accept: the requested response content type,
        either ``application/json`` (default) or ``application/yaml``
    :resheader Content-Type: this depends on :http:header:`Accept` header
        of request

    :>json object author: information about the author of the revision
    :>json object committer: information about the committer of the revision
    :>json string committer_date: ISO representation of the commit date
        (in UTC)
    :>json string date: ISO representation of the revision date (in UTC)
    :>json string directory: the unique identifier that revision points to
    :>json string directory_url: link to
        :http:get:`/api/1/directory/(sha1_git)/[(path)/]` to get information
        about the directory associated to the revision
    :>json string id: the revision unique identifier
    :>json boolean merge: whether or not the revision corresponds to a merge
        commit
    :>json string message: the message associated to the revision
    :>json array parents: the parents of the revision, i.e. the previous
        revisions that head directly to it, each entry of that array contains
        an unique parent revision identifier but also a link to
        :http:get:`/api/1/revision/(sha1_git)/` to get more information
        about it
    :>json string type: the type of the revision

    **Allowed HTTP Methods:** :http:method:`get`, :http:method:`head`

    :statuscode 200: no error
    :statuscode 400: an invalid **sha1_git** value has been provided
    :statuscode 404: requested revision can not be found in the archive

    **Request:**

    .. parsed-literal::

        :swh_web_api:`revision/aafb16d69fd30ff58afdd69036a26047f3aebdc6/`
"""


exception_http_code = {
    BadInputExc: 400,
    ForbiddenExc: 403,
    NotFoundExc: 404,
    Exception: 500,
    StorageAPIError: 503,
    StorageDBError: 503,
}


def test_apidoc_nodoc_failure():
    with pytest.raises(Exception):
        @api_doc('/my/nodoc/url/')
        def apidoc_nodoc_tester(request, arga=0, argb=0):
            return Response(arga + argb)


@api_route(r'/some/(?P<myarg>[0-9]+)/(?P<myotherarg>[0-9]+)/',
           'some-doc-route')
@api_doc('/some/doc/route/')
def apidoc_route(request, myarg, myotherarg, akw=0):
    """
    Sample doc
    """
    return {'result': int(myarg) + int(myotherarg) + akw}

# remove deprecation warnings related to docutils
@pytest.mark.filterwarnings(
    'ignore:.*U.*mode is deprecated:DeprecationWarning')
def test_apidoc_route_doc(client):
    rv = client.get('/api/1/some/doc/route/', HTTP_ACCEPT='text/html')

    assert rv.status_code == 200, rv.content
    assert_template_used(rv, 'api/apidoc.html')


def test_apidoc_route_fn(api_client):
    rv = api_client.get('/api/1/some/1/1/')

    assert rv.status_code == 200, rv.data


@api_route(r'/test/error/(?P<exc_name>.+)/', 'test-error')
@api_doc('/test/error/')
def apidoc_test_error_route(request, exc_name):
    """
    Sample doc
    """
    for e in exception_http_code.keys():
        if e.__name__ == exc_name:
            raise e('Error')


def test_apidoc_error(api_client):
    for exc, code in exception_http_code.items():
        rv = api_client.get('/api/1/test/error/%s/' % exc.__name__)

        assert rv.status_code == code, rv.data


@api_route(r'/some/full/(?P<myarg>[0-9]+)/(?P<myotherarg>[0-9]+)/',
           'some-complete-doc-route')
@api_doc('/some/complete/doc/route/')
def apidoc_full_stack(request, myarg, myotherarg, akw=0):
    """
    Sample doc
    """
    return {'result': int(myarg) + int(myotherarg) + akw}


# remove deprecation warnings related to docutils
@pytest.mark.filterwarnings(
    'ignore:.*U.*mode is deprecated:DeprecationWarning')
def test_apidoc_full_stack_doc(client):
    rv = client.get('/api/1/some/complete/doc/route/', HTTP_ACCEPT='text/html')
    assert rv.status_code == 200, rv.content
    assert_template_used(rv, 'api/apidoc.html')


def test_apidoc_full_stack_fn(api_client):
    rv = api_client.get('/api/1/some/full/1/1/')

    assert rv.status_code == 200, rv.data


def test_api_doc_parse_httpdomain():
    doc_data = {
        'description': '',
        'urls': [],
        'args': [],
        'params': [],
        'resheaders': [],
        'reqheaders': [],
        'return_type': '',
        'returns': [],
        'status_codes': [],
        'examples': []
    }

    _parse_httpdomain_doc(httpdomain_doc, doc_data)

    expected_urls = [{
        'rule': '/api/1/revision/ **\\(sha1_git\\)** /',
        'methods': ['GET', 'HEAD']
    }]

    assert 'urls' in doc_data
    assert doc_data['urls'] == expected_urls

    expected_description = ('Get information about a revision in the archive. '
                            'Revisions are identified by **sha1** checksums, '
                            'compatible with Git commit identifiers. See '
                            '**swh.model.identifiers.revision_identifier** in '
                            'our data model module for details about how they '
                            'are computed.')

    assert 'description' in doc_data
    assert doc_data['description'] == expected_description

    expected_args = [{
        'name': 'sha1_git',
        'type': 'string',
        'doc': ('hexadecimal representation of the revision '
                '**sha1_git** identifier')
    }]

    assert 'args' in doc_data
    assert doc_data['args'] == expected_args

    expected_params = []
    assert 'params' in doc_data
    assert doc_data['params'] == expected_params

    expected_reqheaders = [{
        'doc': ('the requested response content type, either '
                '``application/json``  or ``application/yaml``'),
        'name': 'Accept'
    }]

    assert 'reqheaders' in doc_data
    assert doc_data['reqheaders'] == expected_reqheaders

    expected_resheaders = [{
        'doc': 'this depends on **Accept** header of request',
        'name': 'Content-Type'
    }]

    assert 'resheaders' in doc_data
    assert doc_data['resheaders'] == expected_resheaders

    expected_statuscodes = [
        {
            'code': '200',
            'doc': 'no error'
        },
        {
            'code': '400',
            'doc': 'an invalid **sha1_git** value has been provided'
        },
        {
            'code': '404',
            'doc': 'requested revision can not be found in the archive'
        }
    ]

    assert 'status_codes' in doc_data
    assert doc_data['status_codes'] == expected_statuscodes

    expected_return_type = 'object'

    assert 'return_type' in doc_data
    assert doc_data['return_type'] in expected_return_type

    expected_returns = [
        {
            'name': 'author',
            'type': 'object',
            'doc': 'information about the author of the revision'
        },
        {
            'name': 'committer',
            'type': 'object',
            'doc': 'information about the committer of the revision'
        },
        {
            'name': 'committer_date',
            'type': 'string',
            'doc': 'ISO representation of the commit date (in UTC)'
        },
        {
            'name': 'date',
            'type': 'string',
            'doc': 'ISO representation of the revision date (in UTC)'
        },
        {
            'name': 'directory',
            'type': 'string',
            'doc': 'the unique identifier that revision points to'
        },
        {
            'name': 'directory_url',
            'type': 'string',
            'doc': ('link to `</api/1/directory/>`_ to get information about '
                    'the directory associated to the revision')
        },
        {
            'name': 'id',
            'type': 'string',
            'doc': 'the revision unique identifier'
        },
        {
            'name': 'merge',
            'type': 'boolean',
            'doc': 'whether or not the revision corresponds to a merge commit'
        },
        {
            'name': 'message',
            'type': 'string',
            'doc': 'the message associated to the revision'
        },
        {
            'name': 'parents',
            'type': 'array',
            'doc': ('the parents of the revision, i.e. the previous revisions '
                    'that head directly to it, each entry of that array '
                    'contains an unique parent revision identifier but also a '
                    'link to `</api/1/revision/>`_ to get more information '
                    'about it')
        },
        {
            'name': 'type',
            'type': 'string',
            'doc': 'the type of the revision'
        }
    ]

    assert 'returns' in doc_data
    assert doc_data['returns'] == expected_returns

    expected_examples = [
        '/api/1/revision/aafb16d69fd30ff58afdd69036a26047f3aebdc6/'
    ]

    assert 'examples' in doc_data
    assert doc_data['examples'] == expected_examples



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

from swh.web.api.apiresponse import (
    compute_link_header, transform, make_api_response,
    filter_by_fields
)


def test_compute_link_header():
    next_link = '/api/endpoint/next'
    prev_link = '/api/endpoint/prev'
    rv = {
        'headers': {'link-next': next_link, 'link-prev': prev_link},
        'results': [1, 2, 3]
    }
    options = {}

    headers = compute_link_header(rv, options)

    assert headers == {
        'Link': (f'<{next_link}>; rel="next",'
                 f'<{prev_link}>; rel="previous"')
    }


def test_compute_link_header_nothing_changed():
    rv = {}
    options = {}

    headers = compute_link_header(rv, options)

    assert headers == {}


def test_compute_link_header_nothing_changed_2():
    rv = {'headers': {}}
    options = {}

    headers = compute_link_header(rv, options)

    assert headers == {}


def test_transform_only_return_results_1():
    rv = {'results': {'some-key': 'some-value'}}
    assert transform(rv) == {'some-key': 'some-value'}


def test_transform_only_return_results_2():
    rv = {'headers': {'something': 'do changes'},
          'results': {'some-key': 'some-value'}}
    assert transform(rv) == {'some-key': 'some-value'}


def test_transform_do_remove_headers():
    rv = {'headers': {'something': 'do changes'},
          'some-key': 'some-value'}
    assert transform(rv) == {'some-key': 'some-value'}


def test_transform_do_nothing():
    rv = {'some-key': 'some-value'}
    assert transform(rv) == {'some-key': 'some-value'}


def test_swh_multi_response_mimetype(mocker, api_request_factory):
    mock_shorten_path = mocker.patch('swh.web.api.apiresponse.shorten_path')
    mock_filter = mocker.patch('swh.web.api.apiresponse.filter_by_fields')
    mock_json = mocker.patch('swh.web.api.apiresponse.json')

    data = {
        'data': [12, 34],
        'id': 'adc83b19e793491b1c6ea0fd8b46cd9f32e592fc'
    }

    mock_filter.return_value = data
    mock_shorten_path.return_value = 'my_short_path'

    accepted_response_formats = {'html': 'text/html',
                                 'yaml': 'application/yaml',
                                 'json': 'application/json'}

    for format in accepted_response_formats:

        request = api_request_factory.get('/api/test/path/')

        mime_type = accepted_response_formats[format]
        setattr(request, 'accepted_media_type', mime_type)

        if mime_type == 'text/html':

            expected_data = {
                'response_data': json.dumps(data),
                'headers_data': {},
                'heading': 'my_short_path',
                'status_code': 200
            }

            mock_json.dumps.return_value = json.dumps(data)
        else:
            expected_data = data

        rv = make_api_response(request, data)

        mock_filter.assert_called_with(request, data)

        assert rv.status_code == 200, rv.data
        assert rv.data == expected_data
        if mime_type == 'text/html':
            assert rv.template_name == 'api/apidoc.html'


def test_swh_filter_renderer_do_nothing(api_request_factory):
    input_data = {'a': 'some-data'}

    request = api_request_factory.get('/api/test/path/', data={})
    setattr(request, 'query_params', request.GET)

    actual_data = filter_by_fields(request, input_data)

    assert actual_data == input_data


def test_swh_filter_renderer_do_filter(mocker, api_request_factory):
    mock_ffk = mocker.patch('swh.web.api.apiresponse.utils.filter_field_keys')
    mock_ffk.return_value = {'a': 'some-data'}

    request = api_request_factory.get('/api/test/path/',
                                      data={'fields': 'a,c'})
    setattr(request, 'query_params', request.GET)

    input_data = {'a': 'some-data',
                  'b': 'some-other-data'}

    actual_data = filter_by_fields(request, input_data)

    assert actual_data == {'a': 'some-data'}

    mock_ffk.assert_called_once_with(input_data, {'a', 'c'})



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.storage.exc import StorageDBError, StorageAPIError

from swh.web.common.exc import BadInputExc
from swh.web.common.utils import reverse


def test_api_1_stat_counters_raise_error(api_client, mocker):
    mock_service = mocker.patch('swh.web.api.views.stat.service')
    mock_service.stat_counters.side_effect = BadInputExc(
        'voluntary error to check the bad request middleware.')

    url = reverse('api-1-stat-counters')
    rv = api_client.get(url)

    assert rv.status_code == 400, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'BadInputExc',
        'reason': 'voluntary error to check the bad request middleware.'}


def test_api_1_stat_counters_raise_from_db(api_client, mocker):
    mock_service = mocker.patch('swh.web.api.views.stat.service')
    mock_service.stat_counters.side_effect = StorageDBError(
        'Storage exploded! Will be back online shortly!')

    url = reverse('api-1-stat-counters')
    rv = api_client.get(url)

    assert rv.status_code == 503, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'StorageDBError',
        'reason':
        'An unexpected error occurred in the backend: '
        'Storage exploded! Will be back online shortly!'}


def test_api_1_stat_counters_raise_from_api(api_client, mocker):
    mock_service = mocker.patch('swh.web.api.views.stat.service')
    mock_service.stat_counters.side_effect = StorageAPIError(
        'Storage API dropped dead! Will resurrect from its ashes asap!'
    )

    url = reverse('api-1-stat-counters')
    rv = api_client.get(url)

    assert rv.status_code == 503, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'StorageAPIError',
        'reason':
        'An unexpected error occurred in the api backend: '
        'Storage API dropped dead! Will resurrect from its ashes asap!'
    }


def test_api_1_stat_counters(api_client, archive_data):
    url = reverse('api-1-stat-counters')

    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == archive_data.stat_counters()



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from datetime import datetime, timedelta
from django.utils import timezone

from swh.web.common.utils import reverse
from swh.web.common.models import (
    SaveUnauthorizedOrigin, SaveOriginRequest,
    SAVE_REQUEST_ACCEPTED, SAVE_REQUEST_REJECTED,
    SAVE_REQUEST_PENDING
)
from swh.web.common.models import (
    SAVE_TASK_NOT_CREATED, SAVE_TASK_NOT_YET_SCHEDULED,
    SAVE_TASK_SCHEDULED, SAVE_TASK_FAILED, SAVE_TASK_SUCCEED
)

pytestmark = pytest.mark.django_db


@pytest.fixture(autouse=True)
def populated_db():
    SaveUnauthorizedOrigin.objects.create(
        url='https://github.com/user/illegal_repo')
    SaveUnauthorizedOrigin.objects.create(
        url='https://gitlab.com/user_to_exclude')


def test_invalid_visit_type(api_client):
    url = reverse('api-1-save-origin',
                  url_args={'visit_type': 'foo',
                            'origin_url': 'https://github.com/torvalds/linux'})

    response = api_client.post(url)
    assert response.status_code == 400


def test_invalid_origin_url(api_client):
    url = reverse('api-1-save-origin',
                  url_args={'visit_type': 'git',
                            'origin_url': 'bar'})

    response = api_client.post(url)
    assert response.status_code == 400


def check_created_save_request_status(api_client, mocker, origin_url,
                                      scheduler_task_status,
                                      expected_request_status,
                                      expected_task_status=None,
                                      visit_date=None):

    mock_scheduler = mocker.patch('swh.web.common.origin_save.scheduler')
    if not scheduler_task_status:
        mock_scheduler.get_tasks.return_value = []
    else:
        mock_scheduler.get_tasks.return_value = [{
            'priority': 'high',
            'policy': 'oneshot',
            'type': 'load-git',
            'arguments': {
                'kwargs': {
                    'repo_url': origin_url
                },
                'args': []
            },
            'status': scheduler_task_status,
            'id': 1,
        }]

    mock_scheduler.create_tasks.return_value = [{
        'priority': 'high',
        'policy': 'oneshot',
        'type': 'load-git',
        'arguments': {
            'kwargs': {
                'repo_url': origin_url
            },
            'args': []
        },
        'status': 'next_run_not_scheduled',
        'id': 1,
    }]

    url = reverse('api-1-save-origin',
                  url_args={'visit_type': 'git',
                            'origin_url': origin_url})

    mock_visit_date = mocker.patch(('swh.web.common.origin_save.'
                                    '_get_visit_info_for_save_request'))
    mock_visit_date.return_value = (visit_date, None)
    response = api_client.post(url)

    if expected_request_status != SAVE_REQUEST_REJECTED:
        assert response.status_code == 200, response.data
        assert (response.data['save_request_status'] ==
                expected_request_status)
        assert response.data['save_task_status'] == expected_task_status
    else:
        assert response.status_code == 403, response.data


def check_save_request_status(api_client, mocker, origin_url,
                              expected_request_status,
                              expected_task_status,
                              scheduler_task_status='next_run_not_scheduled',
                              visit_date=None):
    mock_scheduler = mocker.patch('swh.web.common.origin_save.scheduler')
    mock_scheduler.get_tasks.return_value = [{
        'priority': 'high',
        'policy': 'oneshot',
        'type': 'load-git',
        'arguments': {
            'kwargs': {
                'repo_url': origin_url
            },
            'args': []
        },
        'status': scheduler_task_status,
        'id': 1,
    }]

    url = reverse('api-1-save-origin',
                  url_args={'visit_type': 'git',
                            'origin_url': origin_url})

    mock_visit_date = mocker.patch(('swh.web.common.origin_save.'
                                    '_get_visit_info_for_save_request'))
    mock_visit_date.return_value = (visit_date, None)
    response = api_client.get(url)
    assert response.status_code == 200, response.data
    save_request_data = response.data[0]

    assert (save_request_data['save_request_status'] ==
            expected_request_status)
    assert save_request_data['save_task_status'] == expected_task_status

    # Check that save task status is still available when
    # the scheduler task has been archived
    mock_scheduler.get_tasks.return_value = []
    response = api_client.get(url)
    assert response.status_code == 200
    save_request_data = response.data[0]
    assert save_request_data['save_task_status'] == expected_task_status


def test_save_request_rejected(api_client, mocker):
    origin_url = 'https://github.com/user/illegal_repo'
    check_created_save_request_status(api_client, mocker, origin_url,
                                      None, SAVE_REQUEST_REJECTED)
    check_save_request_status(api_client, mocker, origin_url,
                              SAVE_REQUEST_REJECTED, SAVE_TASK_NOT_CREATED)


def test_save_request_pending(api_client, mocker):
    origin_url = 'https://unkwownforge.com/user/repo'
    check_created_save_request_status(api_client, mocker,
                                      origin_url, None, SAVE_REQUEST_PENDING,
                                      SAVE_TASK_NOT_CREATED)
    check_save_request_status(api_client, mocker, origin_url,
                              SAVE_REQUEST_PENDING, SAVE_TASK_NOT_CREATED)


def test_save_request_succeed(api_client, mocker):
    origin_url = 'https://github.com/Kitware/CMake'
    check_created_save_request_status(api_client, mocker, origin_url,
                                      None, SAVE_REQUEST_ACCEPTED,
                                      SAVE_TASK_NOT_YET_SCHEDULED)
    check_save_request_status(api_client, mocker, origin_url,
                              SAVE_REQUEST_ACCEPTED, SAVE_TASK_SCHEDULED,
                              scheduler_task_status='next_run_scheduled')
    check_save_request_status(api_client, mocker, origin_url,
                              SAVE_REQUEST_ACCEPTED, SAVE_TASK_SUCCEED,
                              scheduler_task_status='completed',
                              visit_date=None)
    visit_date = datetime.now(tz=timezone.utc) + timedelta(hours=1)
    check_save_request_status(api_client, mocker, origin_url,
                              SAVE_REQUEST_ACCEPTED, SAVE_TASK_SUCCEED,
                              scheduler_task_status='completed',
                              visit_date=visit_date)


def test_save_request_failed(api_client, mocker):
    origin_url = 'https://gitlab.com/inkscape/inkscape'
    check_created_save_request_status(api_client, mocker, origin_url,
                                      None, SAVE_REQUEST_ACCEPTED,
                                      SAVE_TASK_NOT_YET_SCHEDULED)
    check_save_request_status(api_client, mocker, origin_url,
                              SAVE_REQUEST_ACCEPTED, SAVE_TASK_SCHEDULED,
                              scheduler_task_status='next_run_scheduled')
    check_save_request_status(api_client, mocker, origin_url,
                              SAVE_REQUEST_ACCEPTED, SAVE_TASK_FAILED,
                              scheduler_task_status='disabled')


def test_create_save_request_only_when_needed(api_client, mocker):
    origin_url = 'https://github.com/webpack/webpack'
    SaveOriginRequest.objects.create(visit_type='git', origin_url=origin_url,
                                     status=SAVE_REQUEST_ACCEPTED,
                                     loading_task_id=56)

    check_created_save_request_status(api_client, mocker, origin_url,
                                      'next_run_not_scheduled',
                                      SAVE_REQUEST_ACCEPTED,
                                      SAVE_TASK_NOT_YET_SCHEDULED)

    sors = list(SaveOriginRequest.objects.filter(visit_type='git',
                                                 origin_url=origin_url))
    assert len(sors) == 1

    check_created_save_request_status(api_client, mocker, origin_url,
                                      'next_run_scheduled',
                                      SAVE_REQUEST_ACCEPTED,
                                      SAVE_TASK_SCHEDULED)
    sors = list(SaveOriginRequest.objects.filter(visit_type='git',
                                                 origin_url=origin_url))
    assert len(sors) == 1

    visit_date = datetime.now(tz=timezone.utc) + timedelta(hours=1)
    check_created_save_request_status(api_client, mocker, origin_url,
                                      'completed', SAVE_REQUEST_ACCEPTED,
                                      SAVE_TASK_NOT_YET_SCHEDULED,
                                      visit_date=visit_date)
    sors = list(SaveOriginRequest.objects.filter(visit_type='git',
                                                 origin_url=origin_url))
    assert len(sors) == 2

    check_created_save_request_status(api_client, mocker, origin_url,
                                      'disabled', SAVE_REQUEST_ACCEPTED,
                                      SAVE_TASK_NOT_YET_SCHEDULED)
    sors = list(SaveOriginRequest.objects.filter(visit_type='git',
                                                 origin_url=origin_url))
    assert len(sors) == 3


def test_get_save_requests_unknown_origin(api_client):
    unknown_origin_url = 'https://gitlab.com/foo/bar'
    url = reverse('api-1-save-origin',
                  url_args={'visit_type': 'git',
                            'origin_url': unknown_origin_url})
    response = api_client.get(url)
    assert response.status_code == 404
    assert response.data == {
        'exception': 'NotFoundExc',
        'reason': ('No save requests found for visit of type '
                   'git on origin with url %s.') % unknown_origin_url
    }



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from swh.web.common import highlightjs


def test_get_hljs_language_from_mime_type():
    lang = highlightjs.get_hljs_language_from_mime_type('text/plain')
    assert lang is None

    lang = highlightjs.get_hljs_language_from_mime_type('text/x-c')
    assert lang == 'cpp'

    lang = highlightjs.get_hljs_language_from_mime_type('text/x-c++')
    assert lang == 'cpp'

    lang = highlightjs.get_hljs_language_from_mime_type('text/x-perl')
    assert lang == 'perl'

    lang = highlightjs.get_hljs_language_from_mime_type('text/x-python')
    assert lang == 'python'

    lang = highlightjs.get_hljs_language_from_mime_type('text/x-msdos-batch')
    assert lang == 'dos'

    lang = highlightjs.get_hljs_language_from_mime_type('text/x-tex')
    assert lang == 'tex'

    lang = highlightjs.get_hljs_language_from_mime_type('text/x-lisp')
    assert lang == 'lisp'

    lang = highlightjs.get_hljs_language_from_mime_type('text/x-java')
    assert lang == 'java'

    lang = highlightjs.get_hljs_language_from_mime_type('text/x-makefile')
    assert lang == 'makefile'

    lang = highlightjs.get_hljs_language_from_mime_type('text/x-shellscript')
    assert lang == 'bash'

    lang = highlightjs.get_hljs_language_from_mime_type('image/png')
    assert lang is None


def test_get_hljs_language_from_filename():

    for filename, language in (
        ('foo', None), ('foo.h', 'cpp'), ('foo.c', 'cpp'), ('foo.c.in', 'cpp'),
        ('foo.cpp', 'cpp'), ('foo.pl', 'perl'), ('foo.py', 'python'),
        ('foo.md', 'markdown'), ('foo.js', 'javascript'), ('foo.bat', 'dos'),
        ('foo.json', 'json'), ('foo.yml', 'yaml'), ('foo.ini', 'ini'),
        ('foo.cfg', 'ini'), ('foo.hy', 'hy'), ('foo.lisp', 'lisp'),
        ('foo.java', 'java'), ('foo.sh', 'bash'), ('foo.cmake', 'cmake'),
        ('foo.ml', 'ocaml'), ('foo.mli', 'ocaml'), ('foo.rb', 'ruby'),
        ('foo.jl', 'julia'), ('Makefile', 'makefile'),
        ('CMakeLists.txt', 'cmake')
    ):
        lang = highlightjs.get_hljs_language_from_filename(filename)
        assert lang == language



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from hypothesis import given

from swh.web.common.utils import reverse
from swh.web.tests.data import random_content
from swh.web.tests.strategies import content, contents_with_ctags
from swh.web.tests.conftest import ctags_json_missing, fossology_missing


@given(content())
def test_api_content_filetype(api_client, indexer_data, content):
    indexer_data.content_add_mimetype(content['sha1'])
    url = reverse('api-1-content-filetype',
                  url_args={'q': 'sha1_git:%s' % content['sha1_git']})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    content_url = reverse('api-1-content',
                          url_args={'q': 'sha1:%s' % content['sha1']},
                          request=rv.wsgi_request)
    expected_data = indexer_data.content_get_mimetype(content['sha1'])
    expected_data['content_url'] = content_url
    assert rv.data == expected_data


def test_api_content_filetype_sha_not_found(api_client):
    unknown_content_ = random_content()

    url = reverse('api-1-content-filetype',
                  url_args={'q': 'sha1:%s' % unknown_content_['sha1']})
    rv = api_client.get(url)

    assert rv.status_code == 404, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'No filetype information found for content '
        'sha1:%s.' % unknown_content_['sha1']
    }


@pytest.mark.skip  # Language indexer is disabled
@given(content())
def test_api_content_language(api_client, indexer_data, content):
    indexer_data.content_add_language(content['sha1'])
    url = reverse('api-1-content-language',
                  url_args={'q': 'sha1_git:%s' % content['sha1_git']})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    content_url = reverse('api-1-content',
                          url_args={'q': 'sha1:%s' % content['sha1']},
                          request=rv.wsgi_request)
    expected_data = indexer_data.content_get_language(content['sha1'])
    expected_data['content_url'] = content_url
    assert rv.data == expected_data


def test_api_content_language_sha_not_found(api_client):
    unknown_content_ = random_content()

    url = reverse('api-1-content-language',
                  url_args={'q': 'sha1:%s' % unknown_content_['sha1']})
    rv = api_client.get(url)

    assert rv.status_code == 404, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'No language information found for content '
        'sha1:%s.' % unknown_content_['sha1']
    }


@pytest.mark.skip  # Language indexer is disabled
@pytest.mark.skipif(ctags_json_missing,
                    reason="requires ctags with json output support")
@given(contents_with_ctags())
def test_api_content_symbol(api_client, indexer_data, contents_with_ctags):
    expected_data = {}
    for content_sha1 in contents_with_ctags['sha1s']:
        indexer_data.content_add_ctags(content_sha1)
        for ctag in indexer_data.content_get_ctags(content_sha1):
            if ctag['name'] == contents_with_ctags['symbol_name']:
                expected_data[content_sha1] = ctag
                break
    url = reverse('api-1-content-symbol',
                  url_args={'q': contents_with_ctags['symbol_name']},
                  query_params={'per_page': 100})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    for entry in rv.data:
        content_sha1 = entry['sha1']
        expected_entry = expected_data[content_sha1]
        for key, view_name in (('content_url', 'api-1-content'),
                               ('data_url', 'api-1-content-raw'),
                               ('license_url', 'api-1-content-license'),
                               ('language_url', 'api-1-content-language'),
                               ('filetype_url', 'api-1-content-filetype')):
            expected_entry[key] = reverse(
                view_name, url_args={'q': 'sha1:%s' % content_sha1},
                request=rv.wsgi_request)
        expected_entry['sha1'] = content_sha1
        del expected_entry['id']
        assert entry == expected_entry
    assert 'Link' not in rv

    url = reverse('api-1-content-symbol',
                  url_args={'q': contents_with_ctags['symbol_name']},
                  query_params={'per_page': 2})
    rv = api_client.get(url)

    next_url = reverse('api-1-content-symbol',
                       url_args={'q': contents_with_ctags['symbol_name']},
                       query_params={'last_sha1': rv.data[1]['sha1'],
                                     'per_page': 2},
                       request=rv.wsgi_request),
    assert rv['Link'] == '<%s>; rel="next"' % next_url


def test_api_content_symbol_not_found(api_client):
    url = reverse('api-1-content-symbol', url_args={'q': 'bar'})
    rv = api_client.get(url)

    assert rv.status_code == 404, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'No indexed raw content match expression \'bar\'.'
    }
    assert 'Link' not in rv


@pytest.mark.skipif(ctags_json_missing,
                    reason="requires ctags with json output support")
@given(content())
def test_api_content_ctags(api_client, indexer_data, content):
    indexer_data.content_add_ctags(content['sha1'])
    url = reverse('api-1-content-ctags',
                  url_args={'q': 'sha1_git:%s' % content['sha1_git']})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    content_url = reverse('api-1-content',
                          url_args={'q': 'sha1:%s' % content['sha1']},
                          request=rv.wsgi_request)
    expected_data = list(indexer_data.content_get_ctags(content['sha1']))
    for e in expected_data:
        e['content_url'] = content_url
    assert rv.data == expected_data


@pytest.mark.skipif(fossology_missing,
                    reason="requires fossology-nomossa installed")
@given(content())
def test_api_content_license(api_client, indexer_data, content):
    indexer_data.content_add_license(content['sha1'])
    url = reverse('api-1-content-license',
                  url_args={'q': 'sha1_git:%s' % content['sha1_git']})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    content_url = reverse('api-1-content',
                          url_args={'q': 'sha1:%s' % content['sha1']},
                          request=rv.wsgi_request)
    expected_data = indexer_data.content_get_license(content['sha1'])
    expected_data['content_url'] = content_url
    assert rv.data == expected_data


def test_api_content_license_sha_not_found(api_client):
    unknown_content_ = random_content()

    url = reverse('api-1-content-license',
                  url_args={'q': 'sha1:%s' % unknown_content_['sha1']})
    rv = api_client.get(url)

    assert rv.status_code == 404, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'No license information found for content '
        'sha1:%s.' % unknown_content_['sha1']
    }


@given(content())
def test_api_content_metadata(api_client, archive_data, content):
    url = reverse('api-1-content', {'q': 'sha1:%s' % content['sha1']})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    expected_data = archive_data.content_get_metadata(content['sha1'])
    for key, view_name in (('data_url', 'api-1-content-raw'),
                           ('license_url', 'api-1-content-license'),
                           ('language_url', 'api-1-content-language'),
                           ('filetype_url', 'api-1-content-filetype')):
        expected_data[key] = reverse(
            view_name, url_args={'q': 'sha1:%s' % content['sha1']},
            request=rv.wsgi_request)
    assert rv.data == expected_data


def test_api_content_not_found_as_json(api_client):
    unknown_content_ = random_content()

    url = reverse('api-1-content',
                  url_args={'q': 'sha1:%s' % unknown_content_['sha1']})
    rv = api_client.get(url)
    assert rv.status_code == 404, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'Content with sha1 checksum equals to %s not found!'
        % unknown_content_['sha1']
    }


def test_api_content_not_found_as_yaml(api_client):
    unknown_content_ = random_content()

    url = reverse('api-1-content',
                  url_args={'q': 'sha256:%s' % unknown_content_['sha256']})
    rv = api_client.get(url, HTTP_ACCEPT='application/yaml')

    assert rv.status_code == 404, rv.data
    assert 'application/yaml' in rv['Content-Type']

    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'Content with sha256 checksum equals to %s not found!' %
        unknown_content_['sha256']
    }


def test_api_content_raw_ko_not_found(api_client):
    unknown_content_ = random_content()

    url = reverse('api-1-content-raw',
                  url_args={'q': 'sha1:%s' % unknown_content_['sha1']})
    rv = api_client.get(url)

    assert rv.status_code == 404, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'Content with sha1 checksum equals to %s not found!' %
        unknown_content_['sha1']
    }


@given(content())
def test_api_content_raw_text(api_client, archive_data, content):
    url = reverse('api-1-content-raw',
                  url_args={'q': 'sha1:%s' % content['sha1']})

    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/octet-stream'
    assert rv['Content-disposition'] == \
        'attachment; filename=content_sha1_%s_raw' % content['sha1']
    assert rv['Content-Type'] == 'application/octet-stream'
    expected_data = archive_data.content_get(content['sha1'])
    assert rv.content == expected_data['data']


@given(content())
def test_api_content_raw_text_with_filename(api_client, archive_data, content):
    url = reverse('api-1-content-raw',
                  url_args={'q': 'sha1:%s' % content['sha1']},
                  query_params={'filename': 'filename.txt'})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/octet-stream'
    assert rv['Content-disposition'] == \
        'attachment; filename=filename.txt'
    assert rv['Content-Type'] == 'application/octet-stream'
    expected_data = archive_data.content_get(content['sha1'])
    assert rv.content == expected_data['data']


@given(content())
def test_api_check_content_known(api_client, content):
    url = reverse('api-1-content-known',
                  url_args={'q': content['sha1']})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'

    assert rv.data == {
        'search_res': [
            {
                'found': True,
                'sha1': content['sha1']
            }
        ],
        'search_stats': {'nbfiles': 1, 'pct': 100.0}
    }


@given(content())
def test_api_check_content_known_as_yaml(api_client, content):
    url = reverse('api-1-content-known',
                  url_args={'q': content['sha1']})
    rv = api_client.get(url, HTTP_ACCEPT='application/yaml')

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/yaml'

    assert rv.data == {
        'search_res': [
            {
                'found': True,
                'sha1': content['sha1']
            }
        ],
        'search_stats': {'nbfiles': 1, 'pct': 100.0}
    }


@given(content())
def test_api_check_content_known_post_as_yaml(api_client, content):
    url = reverse('api-1-content-known')
    rv = api_client.post(url, data={'q': content['sha1']},
                         HTTP_ACCEPT='application/yaml')

    assert rv.status_code == 200, rv.data
    assert 'application/yaml' in rv['Content-Type']
    assert rv.data == {
        'search_res': [
            {
                'found': True,
                'sha1': content['sha1']
            }
        ],
        'search_stats': {'nbfiles': 1, 'pct': 100.0}
    }


def test_api_check_content_known_not_found(api_client):
    unknown_content_ = random_content()

    url = reverse('api-1-content-known',
                  url_args={'q': unknown_content_['sha1']})
    rv = api_client.get(url)

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'search_res': [
            {
                'found': False,
                'sha1': unknown_content_['sha1']
            }
        ],
        'search_stats': {'nbfiles': 1, 'pct': 0.0}
    }


@given(content())
def test_api_content_uppercase(api_client, content):
    url = reverse('api-1-content-uppercase-checksum',
                  url_args={'q': content['sha1'].upper()})

    rv = api_client.get(url)
    assert rv.status_code == 302, rv.data

    redirect_url = reverse('api-1-content',
                           url_args={'q': content['sha1']})

    assert rv['location'] == redirect_url



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime
from hypothesis import given

from swh.model.hashutil import hash_to_bytes
from swh.web.common.utils import reverse
from swh.web.tests.data import random_sha1
from swh.web.tests.strategies import (
    release, sha1, content, directory
)


@given(release())
def test_api_release(api_client, archive_data, release):
    url = reverse('api-1-release', url_args={'sha1_git': release})

    rv = api_client.get(url)

    expected_release = archive_data.release_get(release)
    target_revision = expected_release['target']
    target_url = reverse('api-1-revision',
                         url_args={'sha1_git': target_revision},
                         request=rv.wsgi_request)
    expected_release['target_url'] = target_url

    assert rv.status_code == 200, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == expected_release


@given(sha1(), sha1(), sha1(), content(), directory(), release())
def test_api_release_target_type_not_a_revision(api_client, archive_data,
                                                new_rel1, new_rel2,
                                                new_rel3, content,
                                                directory, release):
    for new_rel_id, target_type, target in (
            (new_rel1, 'content', content),
            (new_rel2, 'directory', directory),
            (new_rel3, 'release', release)):

        if target_type == 'content':
            target = target['sha1_git']

        sample_release = {
            'author': {
                'email': b'author@company.org',
                'fullname': b'author <author@company.org>',
                'name': b'author'
            },
            'date': {
                'timestamp': int(datetime.now().timestamp()),
                'offset': 0,
                'negative_utc': False,
            },
            'id': hash_to_bytes(new_rel_id),
            'message': b'sample release message',
            'name': b'sample release',
            'synthetic': False,
            'target': hash_to_bytes(target),
            'target_type': target_type
        }

        archive_data.release_add([sample_release])

        url = reverse('api-1-release', url_args={'sha1_git': new_rel_id})

        rv = api_client.get(url)

        expected_release = archive_data.release_get(new_rel_id)

        if target_type == 'content':
            url_args = {'q': 'sha1_git:%s' % target}
        else:
            url_args = {'sha1_git': target}

        target_url = reverse('api-1-%s' % target_type,
                             url_args=url_args,
                             request=rv.wsgi_request)
        expected_release['target_url'] = target_url

        assert rv.status_code == 200, rv.data
        assert rv['Content-Type'] == 'application/json'
        assert rv.data == expected_release


def test_api_release_not_found(api_client):
    unknown_release_ = random_sha1()

    url = reverse('api-1-release', url_args={'sha1_git': unknown_release_})

    rv = api_client.get(url)

    assert rv.status_code == 404, rv.data
    assert rv['Content-Type'] == 'application/json'
    assert rv.data == {
        'exception': 'NotFoundExc',
        'reason': 'Release with sha1_git %s not found.' % unknown_release_
    }


@given(release())
def test_api_release_uppercase(api_client, release):
    url = reverse('api-1-release-uppercase-checksum',
                  url_args={'sha1_git': release.upper()})

    resp = api_client.get(url)
    assert resp.status_code == 302

    redirect_url = reverse('api-1-release-uppercase-checksum',
                           url_args={'sha1_git': release})

    assert resp['location'] == redirect_url



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import hashlib
import json
import textwrap

from django.http import HttpResponse
from django.shortcuts import render
from django.template.defaultfilters import filesizeformat
from django.utils.html import escape
from django.utils.safestring import mark_safe

from swh.model.identifiers import persistent_identifier
from swh.web.common import service
from swh.web.common.utils import (
    reverse, format_utc_iso_date, gen_path_info, swh_object_icons
)
from swh.web.common.exc import NotFoundExc, handle_view_exception
from swh.web.browse.browseurls import browse_route
from swh.web.browse.utils import (
    gen_link, gen_revision_link, gen_revision_url,
    get_snapshot_context, get_revision_log_url, get_directory_entries,
    gen_directory_link, request_content, prepare_content_for_display,
    content_display_max_size, gen_snapshot_link, get_readme_to_display,
    get_swh_persistent_ids, format_log_entries, gen_person_mail_link
)


def _gen_content_url(revision, query_string, path, snapshot_context):
    if snapshot_context:
        url_args = snapshot_context['url_args']
        url_args['path'] = path
        query_params = snapshot_context['query_params']
        query_params['revision'] = revision['id']
        content_url = reverse('browse-origin-content',
                              url_args=url_args,
                              query_params=query_params)
    else:
        content_path = '%s/%s' % (revision['directory'], path)
        content_url = reverse('browse-content',
                              url_args={'query_string': query_string},
                              query_params={'path': content_path})
    return content_url


def _gen_diff_link(idx, diff_anchor, link_text):
    if idx < _max_displayed_file_diffs:
        return gen_link(diff_anchor, link_text)
    else:
        return link_text


# TODO: put in conf
_max_displayed_file_diffs = 1000


def _gen_revision_changes_list(revision, changes, snapshot_context):
    """
    Returns a HTML string describing the file changes
    introduced in a revision.
    As this string will be displayed in the browse revision view,
    links to adequate file diffs are also generated.

    Args:
        revision (str): hexadecimal representation of a revision identifier
        changes (list): list of file changes in the revision
        snapshot_context (dict): optional origin context used to reverse
            the content urls

    Returns:
        A string to insert in a revision HTML view.

    """
    changes_msg = []
    for i, change in enumerate(changes):
        hasher = hashlib.sha1()
        from_query_string = ''
        to_query_string = ''
        diff_id = 'diff-'
        if change['from']:
            from_query_string = 'sha1_git:' + change['from']['target']
            diff_id += change['from']['target'] + '-' + change['from_path']
        diff_id += '-'
        if change['to']:
            to_query_string = 'sha1_git:' + change['to']['target']
            diff_id += change['to']['target'] + change['to_path']
        change['path'] = change['to_path'] or change['from_path']
        url_args = {'from_query_string': from_query_string,
                    'to_query_string': to_query_string}
        query_params = {'path': change['path']}
        change['diff_url'] = reverse('diff-contents',
                                     url_args=url_args,
                                     query_params=query_params)

        hasher.update(diff_id.encode('utf-8'))
        diff_id = hasher.hexdigest()
        change['id'] = diff_id
        panel_diff_link = '#panel_' + diff_id

        if change['type'] == 'modify':
            change['content_url'] = \
                _gen_content_url(revision, to_query_string,
                                 change['to_path'], snapshot_context)
            changes_msg.append('modified:  %s' %
                               _gen_diff_link(i, panel_diff_link,
                                              change['to_path']))
        elif change['type'] == 'insert':
            change['content_url'] = \
                _gen_content_url(revision, to_query_string,
                                 change['to_path'], snapshot_context)
            changes_msg.append('new file:  %s' %
                               _gen_diff_link(i, panel_diff_link,
                                              change['to_path']))
        elif change['type'] == 'delete':
            parent = service.lookup_revision(revision['parents'][0])
            change['content_url'] = \
                _gen_content_url(parent,
                                 from_query_string,
                                 change['from_path'], snapshot_context)
            changes_msg.append('deleted:   %s' %
                               _gen_diff_link(i, panel_diff_link,
                                              change['from_path']))
        elif change['type'] == 'rename':
            change['content_url'] = \
                _gen_content_url(revision, to_query_string,
                                 change['to_path'], snapshot_context)
            link_text = change['from_path'] + ' &rarr; ' + change['to_path']
            changes_msg.append('renamed:   %s' %
                               _gen_diff_link(i, panel_diff_link, link_text))
    if not changes:
        changes_msg.append('No changes')
    return mark_safe('\n'.join(changes_msg))


@browse_route(r'revision/(?P<sha1_git>[0-9a-f]+)/diff/',
              view_name='diff-revision',
              checksum_args=['sha1_git'])
def _revision_diff(request, sha1_git):
    """
    Browse internal endpoint to compute revision diff
    """
    try:
        revision = service.lookup_revision(sha1_git)
        snapshot_context = None
        origin_url = request.GET.get('origin_url', None)
        if not origin_url:
            origin_url = request.GET.get('origin', None)
        timestamp = request.GET.get('timestamp', None)
        visit_id = request.GET.get('visit_id', None)
        if origin_url:
            snapshot_context = get_snapshot_context(
                origin_url=origin_url, timestamp=timestamp, visit_id=visit_id)
    except Exception as exc:
        return handle_view_exception(request, exc)

    changes = service.diff_revision(sha1_git)
    changes_msg = _gen_revision_changes_list(revision, changes,
                                             snapshot_context)

    diff_data = {
        'total_nb_changes': len(changes),
        'changes': changes[:_max_displayed_file_diffs],
        'changes_msg': changes_msg
    }
    diff_data_json = json.dumps(diff_data, separators=(',', ': '))
    return HttpResponse(diff_data_json, content_type='application/json')


NB_LOG_ENTRIES = 100


@browse_route(r'revision/(?P<sha1_git>[0-9a-f]+)/log/',
              view_name='browse-revision-log',
              checksum_args=['sha1_git'])
def revision_log_browse(request, sha1_git):
    """
    Django view that produces an HTML display of the history
    log for a revision identified by its id.

    The url that points to it is :http:get:`/browse/revision/(sha1_git)/log/`
    """
    try:
        per_page = int(request.GET.get('per_page', NB_LOG_ENTRIES))
        offset = int(request.GET.get('offset', 0))
        revs_ordering = request.GET.get('revs_ordering', 'committer_date')
        session_key = 'rev_%s_log_ordering_%s' % (sha1_git, revs_ordering)
        rev_log_session = request.session.get(session_key, None)
        rev_log = []
        revs_walker_state = None
        if rev_log_session:
            rev_log = rev_log_session['rev_log']
            revs_walker_state = rev_log_session['revs_walker_state']

        if len(rev_log) < offset+per_page:
            revs_walker = \
                service.get_revisions_walker(revs_ordering, sha1_git,
                                             max_revs=offset+per_page+1,
                                             state=revs_walker_state)

            rev_log += [rev['id'] for rev in revs_walker]
            revs_walker_state = revs_walker.export_state()

        revs = rev_log[offset:offset+per_page]
        revision_log = service.lookup_revision_multiple(revs)

        request.session[session_key] = {
            'rev_log': rev_log,
            'revs_walker_state': revs_walker_state
        }
    except Exception as exc:
        return handle_view_exception(request, exc)

    revs_ordering = request.GET.get('revs_ordering', '')

    prev_log_url = None
    if len(rev_log) > offset + per_page:
        prev_log_url = reverse('browse-revision-log',
                               url_args={'sha1_git': sha1_git},
                               query_params={'per_page': per_page,
                                             'offset': offset + per_page,
                                             'revs_ordering': revs_ordering})

    next_log_url = None
    if offset != 0:
        next_log_url = reverse('browse-revision-log',
                               url_args={'sha1_git': sha1_git},
                               query_params={'per_page': per_page,
                                             'offset': offset - per_page,
                                             'revs_ordering': revs_ordering})

    revision_log_data = format_log_entries(revision_log, per_page)

    swh_rev_id = persistent_identifier('revision', sha1_git)

    return render(request, 'browse/revision-log.html',
                  {'heading': 'Revision history',
                   'swh_object_id': swh_rev_id,
                   'swh_object_name': 'Revisions history',
                   'swh_object_metadata': None,
                   'revision_log': revision_log_data,
                   'revs_ordering': revs_ordering,
                   'next_log_url': next_log_url,
                   'prev_log_url': prev_log_url,
                   'breadcrumbs': None,
                   'top_right_link': None,
                   'snapshot_context': None,
                   'vault_cooking': None,
                   'show_actions_menu': True,
                   'swh_ids': None})


@browse_route(r'revision/(?P<sha1_git>[0-9a-f]+)/',
              r'revision/(?P<sha1_git>[0-9a-f]+)/(?P<extra_path>.+)/',
              view_name='browse-revision',
              checksum_args=['sha1_git'])
def revision_browse(request, sha1_git, extra_path=None):
    """
    Django view that produces an HTML display of a revision
    identified by its id.

    The url that points to it is :http:get:`/browse/revision/(sha1_git)/`.
    """
    try:
        revision = service.lookup_revision(sha1_git)
        origin_info = None
        snapshot_context = None
        origin_url = request.GET.get('origin_url', None)
        if not origin_url:
            origin_url = request.GET.get('origin', None)
        timestamp = request.GET.get('timestamp', None)
        visit_id = request.GET.get('visit_id', None)
        snapshot_id = request.GET.get('snapshot_id', None)
        path = request.GET.get('path', None)
        dir_id = None
        dirs, files = None, None
        content_data = None
        if origin_url:
            try:
                snapshot_context = get_snapshot_context(
                    origin_url=origin_url, timestamp=timestamp,
                    visit_id=visit_id)
            except NotFoundExc:
                raw_rev_url = reverse('browse-revision',
                                      url_args={'sha1_git': sha1_git})
                error_message = \
                    ('The Software Heritage archive has a revision '
                     'with the hash you provided but the origin '
                     'mentioned in your request appears broken: %s. '
                     'Please check the URL and try again.\n\n'
                     'Nevertheless, you can still browse the revision '
                     'without origin information: %s'
                        % (gen_link(origin_url), gen_link(raw_rev_url)))
                raise NotFoundExc(error_message)
            origin_info = snapshot_context['origin_info']
            snapshot_id = snapshot_context['snapshot_id']
        elif snapshot_id:
            snapshot_context = get_snapshot_context(snapshot_id)
        if path:
            file_info = \
                service.lookup_directory_with_path(revision['directory'], path)
            if file_info['type'] == 'dir':
                dir_id = file_info['target']
            else:
                query_string = 'sha1_git:' + file_info['target']
                content_data = request_content(query_string,
                                               raise_if_unavailable=False)
        else:
            dir_id = revision['directory']

        if dir_id:
            path = '' if path is None else (path + '/')
            dirs, files = get_directory_entries(dir_id)
    except Exception as exc:
        return handle_view_exception(request, exc)

    revision_data = {}

    revision_data['author'] = 'None'
    if revision['author']:
        author_link = gen_person_mail_link(revision['author'])
        revision_data['author'] = author_link
    revision_data['committer'] = 'None'
    if revision['committer']:
        committer_link = gen_person_mail_link(revision['committer'])
        revision_data['committer'] = committer_link
    revision_data['committer date'] = \
        format_utc_iso_date(revision['committer_date'])
    revision_data['date'] = format_utc_iso_date(revision['date'])
    revision_data['directory'] = revision['directory']
    if snapshot_context:
        revision_data['snapshot'] = snapshot_id
        browse_snapshot_link = \
            gen_snapshot_link(snapshot_id)
        revision_data['context-independent snapshot'] = browse_snapshot_link

    revision_data['context-independent directory'] = \
        gen_directory_link(revision['directory'])
    revision_data['revision'] = sha1_git
    revision_data['merge'] = revision['merge']
    revision_data['metadata'] = escape(json.dumps(revision['metadata'],
                                       sort_keys=True,
                                       indent=4, separators=(',', ': ')))

    if origin_info:
        revision_data['origin url'] = gen_link(origin_info['url'],
                                               origin_info['url'])
        revision_data['context-independent revision'] = \
            gen_revision_link(sha1_git)

    parents = ''
    for p in revision['parents']:
        parent_link = gen_revision_link(p, link_text=None, link_attrs=None,
                                        snapshot_context=snapshot_context)
        parents += parent_link + '<br/>'

    revision_data['parents'] = mark_safe(parents)
    revision_data['synthetic'] = revision['synthetic']
    revision_data['type'] = revision['type']

    message_lines = ['None']
    if revision['message']:
        message_lines = revision['message'].split('\n')

    parents = []
    for p in revision['parents']:
        parent_url = gen_revision_url(p, snapshot_context)
        parents.append({'id': p, 'url': parent_url})

    path_info = gen_path_info(path)

    query_params = {'snapshot_id': snapshot_id,
                    'origin': origin_url,
                    'timestamp': timestamp,
                    'visit_id': visit_id}

    breadcrumbs = []
    breadcrumbs.append({'name': revision['directory'][:7],
                        'url': reverse('browse-revision',
                                       url_args={'sha1_git': sha1_git},
                                       query_params=query_params)})
    for pi in path_info:
        query_params['path'] = pi['path']
        breadcrumbs.append({'name': pi['name'],
                            'url': reverse('browse-revision',
                                           url_args={'sha1_git': sha1_git},
                                           query_params=query_params)})

    vault_cooking = {
        'directory_context': False,
        'directory_id': None,
        'revision_context': True,
        'revision_id': sha1_git
    }

    swh_objects = [{'type': 'revision',
                    'id': sha1_git}]

    content = None
    content_size = None
    mimetype = None
    language = None
    readme_name = None
    readme_url = None
    readme_html = None
    readmes = {}
    error_code = 200
    error_message = ''
    error_description = ''

    if content_data:
        breadcrumbs[-1]['url'] = None
        content_size = content_data['length']
        mimetype = content_data['mimetype']
        if content_data['raw_data']:
            content_display_data = prepare_content_for_display(
                content_data['raw_data'], content_data['mimetype'], path)
            content = content_display_data['content_data']
            language = content_display_data['language']
            mimetype = content_display_data['mimetype']
        query_params = {}
        if path:
            filename = path_info[-1]['name']
            query_params['filename'] = path_info[-1]['name']
            revision_data['filename'] = filename

        top_right_link = {
            'url': reverse('browse-content-raw',
                           url_args={'query_string': query_string},
                           query_params=query_params),
            'icon': swh_object_icons['content'],
            'text': 'Raw File'
        }

        swh_objects.append({'type': 'content',
                            'id': file_info['target']})

        error_code = content_data['error_code']
        error_message = content_data['error_message']
        error_description = content_data['error_description']
    else:
        for d in dirs:
            if d['type'] == 'rev':
                d['url'] = reverse('browse-revision',
                                   url_args={'sha1_git': d['target']})
            else:
                query_params['path'] = path + d['name']
                d['url'] = reverse('browse-revision',
                                   url_args={'sha1_git': sha1_git},
                                   query_params=query_params)
        for f in files:
            query_params['path'] = path + f['name']
            f['url'] = reverse('browse-revision',
                               url_args={'sha1_git': sha1_git},
                               query_params=query_params)
            if f['length'] is not None:
                f['length'] = filesizeformat(f['length'])
            if f['name'].lower().startswith('readme'):
                readmes[f['name']] = f['checksums']['sha1']

        readme_name, readme_url, readme_html = get_readme_to_display(readmes)

        top_right_link = {
            'url': get_revision_log_url(sha1_git, snapshot_context),
            'icon': swh_object_icons['revisions history'],
            'text': 'History'
        }

        vault_cooking['directory_context'] = True
        vault_cooking['directory_id'] = dir_id

        swh_objects.append({'type': 'directory',
                            'id': dir_id})

    diff_revision_url = reverse('diff-revision',
                                url_args={'sha1_git': sha1_git},
                                query_params={'origin': origin_url,
                                              'timestamp': timestamp,
                                              'visit_id': visit_id})

    if snapshot_id:
        swh_objects.append({'type': 'snapshot',
                            'id': snapshot_id})

    swh_ids = get_swh_persistent_ids(swh_objects, snapshot_context)

    heading = 'Revision - %s - %s' %\
        (sha1_git[:7], textwrap.shorten(message_lines[0], width=70))
    if snapshot_context:
        context_found = 'snapshot: %s' % snapshot_context['snapshot_id']
        if origin_info:
            context_found = 'origin: %s' % origin_info['url']
        heading += ' - %s' % context_found

    return render(request, 'browse/revision.html',
                  {'heading': heading,
                   'swh_object_id': swh_ids[0]['swh_id'],
                   'swh_object_name': 'Revision',
                   'swh_object_metadata': revision_data,
                   'message_header': message_lines[0],
                   'message_body': '\n'.join(message_lines[1:]),
                   'parents': parents,
                   'snapshot_context': snapshot_context,
                   'dirs': dirs,
                   'files': files,
                   'content': content,
                   'content_size': content_size,
                   'max_content_size': content_display_max_size,
                   'mimetype': mimetype,
                   'language': language,
                   'readme_name': readme_name,
                   'readme_url': readme_url,
                   'readme_html': readme_html,
                   'breadcrumbs': breadcrumbs,
                   'top_right_link': top_right_link,
                   'vault_cooking': vault_cooking,
                   'diff_revision_url': diff_revision_url,
                   'show_actions_menu': True,
                   'swh_ids': swh_ids,
                   'error_code': error_code,
                   'error_message': error_message,
                   'error_description': error_description},
                  status=error_code)



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from inspect import cleandoc
import json
import re

from django import template
from django.core.serializers.json import DjangoJSONEncoder
from django.utils.safestring import mark_safe

from docutils.core import publish_parts
from docutils.writers.html4css1 import Writer, HTMLTranslator

import sentry_sdk

from swh.web.common.origin_save import get_savable_visit_types

register = template.Library()


class NoHeaderHTMLTranslator(HTMLTranslator):
    """
    Docutils translator subclass to customize the generation of HTML
    from reST-formatted docstrings
    """
    def __init__(self, document):
        super().__init__(document)
        self.body_prefix = []
        self.body_suffix = []

    def visit_bullet_list(self, node):
        self.context.append((self.compact_simple, self.compact_p))
        self.compact_p = None
        self.compact_simple = self.is_compactable(node)
        self.body.append(self.starttag(node, 'ul', CLASS='docstring'))


DOCSTRING_WRITER = Writer()
DOCSTRING_WRITER.translator_class = NoHeaderHTMLTranslator


@register.filter
def safe_docstring_display(docstring):
    """
    Utility function to htmlize reST-formatted documentation in browsable
    api.
    """
    docstring = cleandoc(docstring)
    return publish_parts(docstring, writer=DOCSTRING_WRITER)['html_body']


@register.filter
def urlize_links_and_mails(text):
    """Utility function for decorating api links in browsable api.

    Args:
        text: whose content matching links should be transformed into
        contextual API or Browse html links.

    Returns
        The text transformed if any link is found.
        The text as is otherwise.

    """
    try:
        if 'href="' not in text:
            text = re.sub(r'(http.*)', r'<a href="\1">\1</a>', text)
            return re.sub(r'([^ <>"]+@[^ <>"]+)',
                          r'<a href="mailto:\1">\1</a>', text)
    except Exception as exc:
        sentry_sdk.capture_exception(exc)

    return text


@register.filter
def urlize_header_links(text):
    """Utility function for decorating headers links in browsable api.

    Args
        text: Text whose content contains Link header value

    Returns:
        The text transformed with html link if any link is found.
        The text as is otherwise.

    """
    links = text.split(',')
    ret = ''
    for i, link in enumerate(links):
        ret += re.sub(r'<(http.*)>', r'<<a href="\1">\1</a>>', link)
        # add one link per line and align them
        if i != len(links) - 1:
            ret += '\n     '
    return ret


@register.filter
def jsonify(obj):
    """Utility function for converting a django template variable
    to JSON in order to use it in script tags.

    Args
        obj: Any django template context variable

    Returns:
        JSON representation of the variable.

    """
    return mark_safe(json.dumps(obj, cls=DjangoJSONEncoder))


@register.filter
def sub(value, arg):
    """Django template filter for subtracting two numbers

    Args:
        value (int/float): the value to subtract from
        arg (int/float): the value to subtract to

    Returns:
        int/float: The subtraction result
    """
    return value - arg


@register.filter
def mul(value, arg):
    """Django template filter for multiplying two numbers

    Args:
        value (int/float): the value to multiply from
        arg (int/float): the value to multiply with

    Returns:
        int/float: The multiplication result
    """
    return value * arg


@register.filter
def key_value(dict, key):
    """Django template filter to get a value in a dictionary.

        Args:
            dict (dict): a dictionary
            key (str): the key to lookup value

        Returns:
            The requested value in the dictionary
    """
    return dict[key]


@register.filter
def visit_type_savable(visit_type):
    """Django template filter to check if a save request can be
    created for a given visit type.

        Args:
            visit_type (str): the type of visit

        Returns:
            If the visit type is saveable or not
    """
    return visit_type in get_savable_visit_types()


@register.filter
def split(value, arg):
    """Django template filter to split a string.

        Args:
            value (str): the string to split
            arg (str): the split separator

        Returns:
            list: the split string parts
    """
    return value.split(arg)



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.web.common.exc import NotFoundExc
from swh.web.api.views import utils


def test_genericapi_lookup_nothing_is_found():

    def test_generic_lookup_fn(sha1, another_unused_arg):
        assert another_unused_arg == 'unused_arg'
        assert sha1 == 'sha1'
        return None

    notfound_msg = 'This will be raised because None is returned.'

    with pytest.raises(NotFoundExc) as e:
        utils.api_lookup(
            test_generic_lookup_fn, 'sha1', 'unused_arg',
            notfound_msg=notfound_msg)

    assert e.match(notfound_msg)


def test_generic_api_map_are_enriched_and_transformed_to_list():

    def test_generic_lookup_fn_1(criteria0, param0, param1):
        assert criteria0 == 'something'
        return map(lambda x: x + 1, [1, 2, 3])

    actual_result = utils.api_lookup(
        test_generic_lookup_fn_1, 'something', 'some param 0',
        'some param 1',
        notfound_msg=('This is not the error message you are looking for. '
                      'Move along.'),
        enrich_fn=lambda x, request: x * 2)

    assert actual_result == [4, 6, 8]


def test_generic_api_list_are_enriched_too():

    def test_generic_lookup_fn_2(crit):
        assert crit == 'something'
        return ['a', 'b', 'c']

    actual_result = utils.api_lookup(
        test_generic_lookup_fn_2, 'something',
        notfound_msg=('Not the error message you are looking for, it is. '
                      'Along, you move!'),
        enrich_fn=lambda x, request: ''. join(['=', x, '=']))

    assert actual_result == ['=a=', '=b=', '=c=']


def test_generic_api_generator_are_enriched_and_returned_as_list():

    def test_generic_lookup_fn_3(crit):
        assert crit == 'crit'
        return (i for i in [4, 5, 6])

    actual_result = utils.api_lookup(
        test_generic_lookup_fn_3, 'crit',
        notfound_msg='Move!',
        enrich_fn=lambda x, request: x - 1)

    assert actual_result == [3, 4, 5]


def test_generic_api_simple_data_are_enriched_and_returned_too():

    def test_generic_lookup_fn_4(crit):
        assert crit == '123'
        return {'a': 10}

    def test_enrich_data(x, request):
        x['a'] = x['a'] * 10
        return x

    actual_result = utils.api_lookup(
        test_generic_lookup_fn_4, '123',
        notfound_msg='Nothing to do',
        enrich_fn=test_enrich_data)

    assert actual_result == {'a': 100}


def test_api_lookup_not_found():
    notfound_msg = 'this is the error message raised as it is None'
    with pytest.raises(NotFoundExc) as e:
        utils.api_lookup(
            lambda x: None, 'something',
            notfound_msg=notfound_msg)

    assert e.match(notfound_msg)


def test_api_lookup_with_result():
    actual_result = utils.api_lookup(
        lambda x: x + '!', 'something',
        notfound_msg='this is the error which won\'t be used here')

    assert actual_result == 'something!'


def test_api_lookup_with_result_as_map():
    actual_result = utils.api_lookup(
        lambda x: map(lambda y: y+1, x), [1, 2, 3],
        notfound_msg='this is the error which won\'t be used here')

    assert actual_result == [2, 3, 4]



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.web.settings.tests import (
    scope1_limiter_rate, scope1_limiter_rate_post,
    scope2_limiter_rate, scope2_limiter_rate_post,
    scope3_limiter_rate, scope3_limiter_rate_post
)

from django.conf.urls import url
from django.test.utils import override_settings

from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.decorators import api_view


from swh.web.common.throttling import SwhWebRateThrottle, throttle_scope


class MockViewScope1(APIView):
    throttle_classes = (SwhWebRateThrottle,)
    throttle_scope = 'scope1'

    def get(self, request):
        return Response('foo_get')

    def post(self, request):
        return Response('foo_post')


@api_view(['GET', 'POST'])
@throttle_scope('scope2')
def mock_view_scope2(request):
    if request.method == 'GET':
        return Response('bar_get')
    elif request.method == 'POST':
        return Response('bar_post')


class MockViewScope3(APIView):
    throttle_classes = (SwhWebRateThrottle,)
    throttle_scope = 'scope3'

    def get(self, request):
        return Response('foo_get')

    def post(self, request):
        return Response('foo_post')


@api_view(['GET', 'POST'])
@throttle_scope('scope3')
def mock_view_scope3(request):
    if request.method == 'GET':
        return Response('bar_get')
    elif request.method == 'POST':
        return Response('bar_post')


urlpatterns = [
    url(r'^scope1_class$', MockViewScope1.as_view()),
    url(r'^scope2_func$', mock_view_scope2),
    url(r'^scope3_class$', MockViewScope3.as_view()),
    url(r'^scope3_func$', mock_view_scope3)
]


def check_response(response, status_code,
                   limit=None, remaining=None):
    assert response.status_code == status_code
    if limit is not None:
        assert response['X-RateLimit-Limit'] == str(limit)
    else:
        assert 'X-RateLimit-Limit' not in response
    if remaining is not None:
        assert response['X-RateLimit-Remaining'] == str(remaining)
    else:
        assert 'X-RateLimit-Remaining' not in response


@override_settings(ROOT_URLCONF=__name__)
def test_scope1_requests_are_throttled(api_client):
    """
    Ensure request rate is limited in scope1
    """
    for i in range(scope1_limiter_rate):
        response = api_client.get('/scope1_class')
        check_response(response, 200, scope1_limiter_rate,
                       scope1_limiter_rate - i - 1)

    response = api_client.get('/scope1_class')
    check_response(response, 429, scope1_limiter_rate, 0)

    for i in range(scope1_limiter_rate_post):
        response = api_client.post('/scope1_class')
        check_response(response, 200, scope1_limiter_rate_post,
                       scope1_limiter_rate_post - i - 1)

    response = api_client.post('/scope1_class')
    check_response(response, 429, scope1_limiter_rate_post, 0)


@override_settings(ROOT_URLCONF=__name__)
def test_scope2_requests_are_throttled(api_client):
    """
    Ensure request rate is limited in scope2
    """
    for i in range(scope2_limiter_rate):
        response = api_client.get('/scope2_func')
        check_response(response, 200, scope2_limiter_rate,
                       scope2_limiter_rate - i - 1)

    response = api_client.get('/scope2_func')
    check_response(response, 429, scope2_limiter_rate, 0)

    for i in range(scope2_limiter_rate_post):
        response = api_client.post('/scope2_func')
        check_response(response, 200, scope2_limiter_rate_post,
                       scope2_limiter_rate_post - i - 1)

    response = api_client.post('/scope2_func')
    check_response(response, 429, scope2_limiter_rate_post, 0)


@override_settings(ROOT_URLCONF=__name__)
def test_scope3_requests_are_throttled_exempted(api_client):
    """
    Ensure request rate is not limited in scope3 as
    requests coming from localhost are exempted from rate limit.
    """
    for _ in range(scope3_limiter_rate+1):
        response = api_client.get('/scope3_class')
        check_response(response, 200)

    for _ in range(scope3_limiter_rate_post+1):
        response = api_client.post('/scope3_class')
        check_response(response, 200)

    for _ in range(scope3_limiter_rate+1):
        response = api_client.get('/scope3_func')
        check_response(response, 200)

    for _ in range(scope3_limiter_rate_post+1):
        response = api_client.post('/scope3_func')
        check_response(response, 200)



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from django.contrib.auth import get_user_model

username = 'admin'
password = 'admin'
email = 'admin@swh-web.org'

User = get_user_model()

if not User.objects.filter(username=username).exists():
    User.objects.create_superuser(username, email, password)
